from pathlib import Path
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
import SimpleITK as sitk
from tifffile import TiffWriter

from wsireg.reg_images import RegImage
from wsireg.reg_transform_seq import RegTransformSeq
from wsireg.utils.im_utils import (
    format_channel_names,
    get_pyramid_info,
    prepare_ome_xml_str,
)
from wsireg.utils.tile_image_transform import (
    add_tile_location_on_moving,
    compute_sub_res,
    determine_moving_tile_padding,
    generate_tile_coords,
    itk_transform_tiles,
    subres_zarr_to_tiles,
    tile_pad_output_size,
    zarr_to_tiles,
)


class OmeTiffWriter:
    x_size: Optional[int] = None
    y_size: Optional[int] = None
    y_spacing: Optional[Union[int, float]] = None
    x_spacing: Optional[Union[int, float]] = None
    tile_size: int = 512
    pyr_levels: Optional[List[Tuple[int, int]]] = None
    n_pyr_levels: Optional[int] = None
    PhysicalSizeY: Optional[Union[int, float]] = None
    PhysicalSizeX: Optional[Union[int, float]] = None
    subifds: Optional[int] = None
    compression: str = "deflate"

    def __init__(
        self,
        reg_image: RegImage,
        reg_transform_seq: Optional[RegTransformSeq] = None,
    ):
        self.reg_image = reg_image
        self.reg_transform_seq = reg_transform_seq

    def _prepare_image_info(
        self,
        image_name,
        reg_transform_seq: Optional[RegTransformSeq] = None,
        write_pyramid=True,
        tile_size=512,
        compression="default",
        by_tile=False,
    ):

        if reg_transform_seq:
            self.x_size, self.y_size = reg_transform_seq.output_size
            self.x_spacing, self.y_spacing = reg_transform_seq.output_spacing
        else:
            self.y_size, self.x_size = (
                (self.reg_image.im_dims[0], self.reg_image.im_dims[1])
                if self.reg_image.is_rgb
                else (self.reg_image.im_dims[1], self.reg_image.im_dims[2])
            )
            self.y_spacing, self.x_spacing = None, None

        self.tile_size = tile_size
        # protect against too large tile size
        while (
            self.y_size / self.tile_size <= 1
            or self.x_size / self.tile_size <= 1
        ):
            self.tile_size = self.tile_size // 2

        if by_tile:
            self.y_size, self.x_size = tile_pad_output_size(
                self.y_size, self.x_size, tile_size=self.tile_size
            )

        self.pyr_levels, _ = get_pyramid_info(
            self.y_size, self.x_size, self.reg_image.n_ch, self.tile_size
        )
        self.n_pyr_levels = len(self.pyr_levels)

        if reg_transform_seq:
            self.PhysicalSizeY = self.y_spacing
            self.PhysicalSizeX = self.x_spacing
        else:
            self.PhysicalSizeY = self.reg_image.image_res
            self.PhysicalSizeX = self.reg_image.image_res

        channel_names = format_channel_names(
            self.reg_image.channel_names, self.reg_image.n_ch
        )

        self.omexml = prepare_ome_xml_str(
            self.y_size,
            self.x_size,
            self.reg_image.n_ch,
            self.reg_image.im_dtype,
            self.reg_image.is_rgb,
            PhysicalSizeX=self.PhysicalSizeX,
            PhysicalSizeY=self.PhysicalSizeY,
            PhysicalSizeXUnit="µm",
            PhysicalSizeYUnit="µm",
            Name=image_name,
            Channel=None if self.reg_image.is_rgb else {"Name": channel_names},
        )

        self.subifds = self.n_pyr_levels - 1 if write_pyramid is True else None

        if compression == "default":
            print("using default compression")
            self.compression = "jpeg" if self.reg_image.is_rgb else "deflate"
        else:
            self.compression = compression

    def write_image_by_plane(
        self,
        image_name,
        output_dir="",
        write_pyramid=True,
        tile_size=512,
        compression="default",
    ):

        output_file_name = str(Path(output_dir) / f"{image_name}.ome.tiff")
        self._prepare_image_info(
            image_name,
            reg_transform_seq=self.reg_transform_seq,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
        )

        rgb_im_data = []

        print(f"saving to {output_file_name}")
        with TiffWriter(output_file_name, bigtiff=True) as tif:
            if self.reg_image.reader == "sitk":
                self.reg_image.read_full_image()
            for channel_idx in range(self.reg_image.n_ch):
                print(f"transforming : {channel_idx}")
                if self.reg_image.reader != "sitk":
                    image = self.reg_image.read_single_channel(channel_idx)
                    image = np.squeeze(image)
                    image = sitk.GetImageFromArray(image)
                    image.SetSpacing(
                        (self.reg_image.image_res, self.reg_image.image_res)
                    )

                if self.reg_transform_seq:
                    image = self.reg_transform_seq.resampler.Execute(image)
                    # image = transform_plane(
                    #     image, final_transform, composite_transform
                    # )
                    print(f"transformed : {channel_idx}")

                if self.reg_image.is_rgb:
                    rgb_im_data.append(image)
                else:
                    print("saving")
                    if isinstance(image, sitk.Image):
                        image = sitk.GetArrayFromImage(image)

                    options = dict(
                        tile=(self.tile_size, self.tile_size),
                        compression=self.compression,
                        photometric="rgb"
                        if self.reg_image.is_rgb
                        else "minisblack",
                        metadata=None,
                    )
                    # write OME-XML to the ImageDescription tag of the first page
                    description = self.omexml if channel_idx == 0 else None
                    # write channel data
                    print(
                        f" writing channel {channel_idx} - shape: {image.shape}"
                    )
                    tif.write(
                        image,
                        subifds=self.subifds,
                        description=description,
                        **options,
                    )

                    if write_pyramid:
                        for pyr_idx in range(1, self.n_pyr_levels):
                            resize_shape = (
                                self.pyr_levels[pyr_idx][0],
                                self.pyr_levels[pyr_idx][1],
                            )
                            image = cv2.resize(
                                image,
                                resize_shape,
                                cv2.INTER_LINEAR,
                            )
                            print(
                                f"pyramid index {pyr_idx} : channel {channel_idx} shape: {image.shape}"
                            )

                            tif.write(image, **options, subfiletype=1)

            if self.reg_image.is_rgb:
                rgb_im_data = sitk.Compose(rgb_im_data)
                rgb_im_data = sitk.GetArrayFromImage(rgb_im_data)

                options = dict(
                    tile=(self.tile_size, self.tile_size),
                    compression=self.compression,
                    photometric="rgb",
                    metadata=None,
                )
                # write OME-XML to the ImageDescription tag of the first page
                description = self.omexml

                # write channel data
                tif.write(
                    rgb_im_data,
                    subifds=self.subifds,
                    description=description,
                    **options,
                )

                print(f"RGB shape: {rgb_im_data.shape}")
                if write_pyramid:
                    for pyr_idx in range(1, self.n_pyr_levels):
                        resize_shape = (
                            self.pyr_levels[pyr_idx][0],
                            self.pyr_levels[pyr_idx][1],
                        )
                        rgb_im_data = cv2.resize(
                            rgb_im_data,
                            resize_shape,
                            cv2.INTER_LINEAR,
                        )
                        tif.write(rgb_im_data, **options, subfiletype=1)
        return output_file_name

    def write_image_by_tile(
        self,
        image_name,
        output_dir="",
        write_pyramid=True,
        tile_size=512,
        compression="default",
        tile_padding=None,
        use_multiprocessing=False,
    ):

        output_file_name = str(Path(output_dir) / f"{image_name}.ome.tiff")

        self._prepare_image_info(
            image_name,
            reg_transform_seq=self.reg_transform_seq,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
            by_tile=True,
        )

        # create per tile coordinate information for fixed
        # and moving
        tile_coordinate_data = generate_tile_coords(
            (self.x_size, self.y_size),
            (self.PhysicalSizeX, self.PhysicalSizeY),
            tile_size=self.tile_size,
        )
        # itk_transforms = [
        #     t.itk_transform for t in self.reg_transform_seq.reg_transforms
        # ]

        tile_coordinate_data = add_tile_location_on_moving(
            self.reg_transform_seq.reg_transforms_itk_order,
            tile_coordinate_data,
            self.reg_image.image_res,
        )

        if tile_padding is None:
            tile_padding = determine_moving_tile_padding(
                self.reg_transform_seq.reg_transforms_itk_order
            )

        print(f"saving to {output_file_name}")
        with TiffWriter(output_file_name, bigtiff=True) as tif:
            options = dict(
                tile=(self.tile_size, self.tile_size),
                compression=self.compression,
                photometric="rgb" if self.reg_image.is_rgb else "minisblack",
                metadata=None,
            )
            if self.reg_image.is_rgb:
                resampled_zarray = itk_transform_tiles(
                    self.reg_image,
                    self.y_size,
                    self.x_size,
                    tile_coordinate_data,
                    self.reg_transform_seq.composite_transform,
                    tile_size=self.tile_size,
                    tile_padding=tile_padding,
                    use_multiprocessing=use_multiprocessing,
                )
                print(
                    f"writing base layer RGB - shape: {resampled_zarray.shape}"
                )
                description = self.omexml
                tif.write(
                    zarr_to_tiles(
                        resampled_zarray,
                        tile_coordinate_data,
                        self.reg_image.is_rgb,
                    ),
                    subifds=self.subifds,
                    description=description,
                    shape=(self.y_size, self.x_size, 3),
                    dtype=resampled_zarray.dtype,
                    **options,
                )
                if write_pyramid:
                    for pyr_idx in range(1, self.n_pyr_levels):

                        resampled_zarray_subres, orig_shape = compute_sub_res(
                            self.reg_image,
                            resampled_zarray,
                            2**pyr_idx,
                            self.tile_size,
                            self.reg_image.is_rgb,
                        )
                        print(f"pyr {pyr_idx} : RGB-shape: {orig_shape}")

                        tif.write(
                            subres_zarr_to_tiles(
                                resampled_zarray_subres,
                                self.tile_size,
                                self.reg_image.is_rgb,
                            ),
                            shape=orig_shape,
                            dtype=resampled_zarray_subres.dtype,
                            **options,
                            subfiletype=1,
                        )
                try:
                    resampled_zarray.store.clear()
                except FileNotFoundError:
                    pass
            else:
                for channel_idx in range(self.reg_image.n_ch):
                    resampled_zarray = itk_transform_tiles(
                        self.reg_image,
                        self.y_size,
                        self.x_size,
                        tile_coordinate_data,
                        self.reg_transform_seq.composite_transform,
                        ch_idx=channel_idx,
                        tile_size=self.tile_size,
                        tile_padding=tile_padding,
                        use_multiprocessing=use_multiprocessing,
                    )

                    description = self.omexml if channel_idx == 0 else None
                    print(
                        f"writing channel {channel_idx} - shape: {resampled_zarray.shape}"
                    )
                    tif.write(
                        zarr_to_tiles(
                            resampled_zarray,
                            tile_coordinate_data,
                            self.reg_image.is_rgb,
                        ),
                        subifds=self.subifds,
                        description=description,
                        shape=(self.y_size, self.x_size),
                        dtype=resampled_zarray.dtype,
                        **options,
                    )
                    if write_pyramid:
                        for pyr_idx in range(1, self.n_pyr_levels):
                            (
                                resampled_zarray_subres,
                                orig_shape,
                            ) = compute_sub_res(
                                self.reg_image,
                                resampled_zarray,
                                2**pyr_idx,
                                self.tile_size,
                                self.reg_image.is_rgb,
                            )

                            tif.write(
                                subres_zarr_to_tiles(
                                    resampled_zarray_subres,
                                    self.tile_size,
                                    self.reg_image.is_rgb,
                                ),
                                shape=orig_shape,
                                dtype=resampled_zarray_subres.dtype,
                                **options,
                                subfiletype=1,
                            )
                    try:
                        resampled_zarray.store.clear()
                    except FileNotFoundError:
                        pass
        return output_file_name



from copy import deepcopy
from typing import Any, Dict, Optional, Union

import itk
import numpy as np
import SimpleITK as sitk

from wsireg.parameter_maps.preprocessing import ImagePreproParams
from wsireg.utils.im_utils import (
    compute_mask_to_bbox,
    contrast_enhance,
    sitk_inv_int,
    sitk_max_int_proj,
    transform_plane,
)
from wsireg.utils.tform_utils import (
    gen_aff_tform_flip,
    gen_rig_to_original,
    gen_rigid_tform_rot,
    gen_rigid_translation,
    prepare_wsireg_transform_data,
)


class RegImage:
    """
    Image container class for reading images and preparing them for registration.
    Does two forms of preprocessing:
    Intensity preprocessing, modifying intensity values for best registration
    Spatial preprocessing, rotating and flipping images prior to registration for best alignment

    After preprocessing, images destined for registration should always be a single channel 2D image

    Attributes
    ----------
    image_filepath : str or np.array
        filepath to image to be processed or a numpy array of the image
    image_res : int or float
        Physical spacing (xy resolution) of the image (units are assumed to be the same for each image in experiment but are not
        defined)
        Does not yet support isotropic resolution
    preprocessing : ImagePreproParams
        preprocessing to apply to the image
    transforms: list
        list of elastix transformation data to apply to an image during preprocessing
    mask : str or np.ndarray
        filepath of a mask image to be processed or a numpy array of the mask
        all values => 1 are considered WITHIN the mask
    """

    def __init__(
        self,
        preprocessing: Optional[
            Union[ImagePreproParams, Dict[str, Any]]
        ] = None,
    ):
        self.image = None
        self.reg_image = None
        self.original_size_transform = None
        if preprocessing:
            if isinstance(preprocessing, ImagePreproParams):
                self.preprocessing = preprocessing
            elif isinstance(preprocessing, dict):
                self.preprocessing = ImagePreproParams(**preprocessing)
        else:
            self.preprocessing = ImagePreproParams()

        self.pre_reg_transforms = None
        self.original_size_transform = None
        self.channel_names = None
        self.channel_colors = None

        return

    def read_mask(self, mask):
        if mask is not None:
            if isinstance(mask, np.ndarray):
                mask = sitk.GetImageFromArray(mask)
            elif isinstance(mask, str):
                mask = sitk.ReadImage(mask)
            elif isinstance(mask, sitk.Image):
                mask = mask

            mask.SetSpacing((self.image_res, self.image_res))

        return mask

    def preprocess_reg_image_intensity(
        self, image, preprocessing: ImagePreproParams
    ):

        if preprocessing.image_type.value == "FL":
            preprocessing.invert_intensity = False
        elif preprocessing.image_type.value == "BF":
            preprocessing.max_int_proj = False
            preprocessing.contrast_enhance = False
            if self.is_rgb:
                preprocessing.invert_intensity = True

        if preprocessing.max_int_proj:
            image = sitk_max_int_proj(image)

        if preprocessing.contrast_enhance:
            image = contrast_enhance(image)

        if preprocessing.invert_intensity:
            image = sitk_inv_int(image)

        if preprocessing.custom_processing:
            for k, v in preprocessing.custom_processing.items():
                print(f"performing preprocessing: {k}")
                image = v(image)

        image.SetSpacing((self.image_res, self.image_res))

        return image

    def preprocess_reg_image_spatial(
        self, image, preprocessing: ImagePreproParams, imported_transforms=None
    ):

        transforms = []
        original_size = image.GetSize()

        if preprocessing.downsampling > 1:
            print(
                "performing downsampling by factor: {}".format(
                    preprocessing.downsampling
                )
            )
            image.SetSpacing((self.image_res, self.image_res))
            image = sitk.Shrink(
                image,
                (preprocessing.downsampling, preprocessing.downsampling),
            )

            if self.mask is not None:
                self.mask.SetSpacing((self.image_res, self.image_res))
                self.mask = sitk.Shrink(
                    self.mask,
                    (
                        preprocessing.downsampling,
                        preprocessing.downsampling,
                    ),
                )

            image_res = image.GetSpacing()[0]
        else:
            image_res = self.image_res

        if float(preprocessing.rot_cc) != 0.0:
            print(f"rotating counter-clockwise {preprocessing.rot_cc}")
            rot_tform = gen_rigid_tform_rot(
                image, image_res, preprocessing.rot_cc
            )
            (
                composite_transform,
                _,
                final_tform,
            ) = prepare_wsireg_transform_data({"initial": [rot_tform]})

            image = transform_plane(image, final_tform, composite_transform)

            if self.mask is not None:
                self.mask.SetSpacing((image_res, image_res))
                self.mask = transform_plane(
                    self.mask, final_tform, composite_transform
                )
            transforms.append(rot_tform)

        if preprocessing.flip:
            print(f"flipping image {preprocessing.flip.value}")

            flip_tform = gen_aff_tform_flip(
                image, image_res, preprocessing.flip.value
            )

            (
                composite_transform,
                _,
                final_tform,
            ) = prepare_wsireg_transform_data({"initial": [flip_tform]})

            image = transform_plane(image, final_tform, composite_transform)

            if self.mask is not None:
                self.mask.SetSpacing((image_res, image_res))
                self.mask = transform_plane(
                    self.mask, final_tform, composite_transform
                )

            transforms.append(flip_tform)

        if self.mask and preprocessing.crop_to_mask_bbox:
            print("computing mask bounding box")
            if preprocessing.mask_bbox is None:
                mask_bbox = compute_mask_to_bbox(self.mask)
                preprocessing.mask_bbox = mask_bbox

        if preprocessing.mask_bbox:

            print("cropping to mask")
            translation_transform = gen_rigid_translation(
                image,
                image_res,
                preprocessing.mask_bbox.X,
                preprocessing.mask_bbox.Y,
                preprocessing.mask_bbox.WIDTH,
                preprocessing.mask_bbox.HEIGHT,
            )

            (
                composite_transform,
                _,
                final_tform,
            ) = prepare_wsireg_transform_data(
                {"initial": [translation_transform]}
            )

            image = transform_plane(image, final_tform, composite_transform)

            self.original_size_transform = gen_rig_to_original(
                original_size, deepcopy(translation_transform)
            )

            if self.mask is not None:
                self.mask.SetSpacing((image_res, image_res))
                self.mask = transform_plane(
                    self.mask, final_tform, composite_transform
                )
            transforms.append(translation_transform)

        return image, transforms

    def preprocess_image(self, reg_image):

        reg_image = self.preprocess_reg_image_intensity(
            reg_image, self.preprocessing
        )

        if reg_image.GetDepth() >= 1:
            raise ValueError(
                "preprocessing did not result in a single image plane\n"
                "multi-channel or 3D image return"
            )

        if reg_image.GetNumberOfComponentsPerPixel() > 1:
            raise ValueError(
                "preprocessing did not result in a single image plane\n"
                "multi-component / RGB(A) image returned"
            )

        reg_image, pre_reg_transforms = self.preprocess_reg_image_spatial(
            reg_image, self.preprocessing, self.pre_reg_transforms
        )

        if len(pre_reg_transforms) > 0:
            self.pre_reg_transforms = pre_reg_transforms

        self.reg_image = reg_image

    def reg_image_sitk_to_itk(self, cast_to_float32=True):
        origin = self.reg_image.GetOrigin()
        spacing = self.reg_image.GetSpacing()
        # direction = image.GetDirection()
        is_vector = self.reg_image.GetNumberOfComponentsPerPixel() > 1
        if cast_to_float32 is True:
            self.reg_image = sitk.Cast(self.reg_image, sitk.sitkFloat32)
            self.reg_image = sitk.GetArrayFromImage(self.reg_image)
        else:
            self.reg_image = sitk.GetArrayFromImage(self.reg_image)

        self.reg_image = itk.GetImageFromArray(
            self.reg_image, is_vector=is_vector
        )
        self.reg_image.SetOrigin(origin)
        self.reg_image.SetSpacing(spacing)

        if self.mask is not None:
            origin = self.mask.GetOrigin()
            spacing = self.mask.GetSpacing()
            # direction = image.GetDirection()
            is_vector = self.mask.GetNumberOfComponentsPerPixel() > 1
            if cast_to_float32 is True:
                self.mask = sitk.Cast(self.mask, sitk.sitkFloat32)
                self.mask = sitk.GetArrayFromImage(self.mask)
            else:
                self.mask = sitk.GetArrayFromImage(self.mask)

            self.mask = itk.GetImageFromArray(self.mask, is_vector=is_vector)
            self.mask.SetOrigin(origin)
            self.mask.SetSpacing(spacing)

            mask_im_type = itk.Image[itk.UC, 2]
            self.mask = itk.binary_threshold_image_filter(
                self.mask,
                lower_threshold=1,
                inside_value=1,
                ttype=(type(self.mask), mask_im_type),
            )



import os

import dask.array as da
import numpy as np
import pytest
import zarr
from tifffile import imread

from wsireg.parameter_maps.preprocessing import ImagePreproParams
from wsireg.utils.im_utils import (
    CziRegImageReader,
    czi_tile_grayscale,
    ensure_dask_array,
    get_sitk_image_info,
    get_tifffile_info,
    grayscale,
    guess_rgb,
    read_preprocess_array,
    sitk_backend,
    tf_get_largest_series,
    tifffile_dask_backend,
    tifffile_zarr_backend,
    zarr_get_base_pyr_layer,
)

# private data logic borrowed from https://github.com/cgohlke/tifffile/tests/test_tifffile.py
HERE = os.path.dirname(__file__)
PRIVATE_DIR = os.path.join(HERE, "private_data")

SKIP_PRIVATE = False
REASON = "private data"

if not os.path.exists(PRIVATE_DIR):
    SKIP_PRIVATE = True


@pytest.mark.usefixtures(
    "disk_im_gry_pyr",
    "disk_im_mch_pyr",
    "disk_im_rgb_pyr",
    "disk_im_gry",
    "disk_im_mch",
    "disk_im_rgb",
)
def test_zarr_get_base_pyr_layer(
    disk_im_gry_pyr,
    disk_im_mch_pyr,
    disk_im_rgb_pyr,
    disk_im_gry,
    disk_im_mch,
    disk_im_rgb,
):
    zarr_store = zarr.open(imread(disk_im_gry_pyr, aszarr=True))
    gry_zarray_from_pyr = zarr_get_base_pyr_layer(zarr_store)

    zarr_store = zarr.open(imread(disk_im_mch_pyr, aszarr=True))
    mch_zarray_from_pyr = zarr_get_base_pyr_layer(zarr_store)

    zarr_store = zarr.open(imread(disk_im_rgb_pyr, aszarr=True))
    rgb_zarray_from_pyr = zarr_get_base_pyr_layer(zarr_store)

    zarr_store = zarr.open(imread(disk_im_gry, aszarr=True))
    gry_zarray_from_flat = zarr_get_base_pyr_layer(zarr_store)

    zarr_store = zarr.open(imread(disk_im_mch, aszarr=True))
    mch_zarray_from_flat = zarr_get_base_pyr_layer(zarr_store)

    zarr_store = zarr.open(imread(disk_im_rgb, aszarr=True))
    rgb_zarray_from_flat = zarr_get_base_pyr_layer(zarr_store)

    assert gry_zarray_from_pyr.shape == (2048, 2048)
    assert mch_zarray_from_pyr.shape == (3, 2048, 2048)
    assert rgb_zarray_from_pyr.shape == (2048, 2048, 3)
    assert gry_zarray_from_flat.shape == (2048, 2048)
    assert mch_zarray_from_flat.shape == (3, 2048, 2048)
    assert rgb_zarray_from_flat.shape == (2048, 2048, 3)


def test_ensure_dask_array():
    np_arr = np.zeros((128, 128, 3), dtype=np.uint8)
    da_arr = da.zeros((128, 128, 3), dtype=np.uint8)
    za_arr = zarr.zeros((128, 128, 3), dtype=np.uint8)

    np_out = ensure_dask_array(np_arr)
    da_out = ensure_dask_array(da_arr)
    za_out = ensure_dask_array(za_arr)

    assert isinstance(np_out, da.Array)
    assert isinstance(da_out, da.Array)
    assert isinstance(za_out, da.Array)


def test_read_preprocess_array():
    da_arr = ensure_dask_array(np.zeros((128, 128, 3), dtype=np.uint8))
    mc_arr = ensure_dask_array(np.zeros((3, 128, 128), dtype=np.uint8))
    gr_arr = ensure_dask_array(np.zeros((128, 128), dtype=np.uint8))

    std_rgb = read_preprocess_array(
        da_arr,
        preprocessing=ImagePreproParams(image_type="BF"),
        force_rgb=None,
    )
    nonstd_rgb = read_preprocess_array(
        mc_arr,
        preprocessing=ImagePreproParams(image_type="BF"),
        force_rgb=True,
    )

    all_ch_mc = read_preprocess_array(
        mc_arr, preprocessing=ImagePreproParams(image_type="FL")
    )

    ch0_ch_mc = read_preprocess_array(
        mc_arr, preprocessing=ImagePreproParams(ch_indices=[0])
    )
    ch01_ch_mc = read_preprocess_array(
        mc_arr, preprocessing=ImagePreproParams(ch_indices=[0, 1])
    )
    ch12_ch_mc = read_preprocess_array(
        mc_arr, preprocessing=ImagePreproParams(ch_indices=[1, 2])
    )
    ch02_ch_mc = read_preprocess_array(
        mc_arr, preprocessing=ImagePreproParams(ch_indices=[0, 2])
    )
    ch1_ch_mc = read_preprocess_array(
        mc_arr, preprocessing=ImagePreproParams(ch_indices=[1])
    )
    ch2_ch_mc = read_preprocess_array(
        mc_arr, preprocessing=ImagePreproParams(ch_indices=[2])
    )

    std_gr = read_preprocess_array(gr_arr, ImagePreproParams(image_type="FL"))
    chsel0_gr = read_preprocess_array(
        gr_arr, preprocessing=ImagePreproParams(ch_indices=[0])
    )
    chsel01_gr = read_preprocess_array(
        gr_arr, preprocessing=ImagePreproParams(ch_indices=[0, 1])
    )

    assert std_rgb.GetNumberOfComponentsPerPixel() == 1
    assert nonstd_rgb.GetNumberOfComponentsPerPixel() == 1
    assert all_ch_mc.GetDepth() == 3
    assert ch0_ch_mc.GetDepth() == 0
    assert ch01_ch_mc.GetDepth() == 2
    assert ch12_ch_mc.GetDepth() == 2
    assert ch02_ch_mc.GetDepth() == 2
    assert ch1_ch_mc.GetDepth() == 0
    assert ch2_ch_mc.GetDepth() == 0
    assert std_gr.GetDepth() == 0
    assert chsel0_gr.GetDepth() == 0
    assert chsel01_gr.GetDepth() == 0
    assert std_rgb.GetSize() == (128, 128)
    assert nonstd_rgb.GetSize() == (128, 128)
    assert all_ch_mc.GetSize() == (128, 128, 3)
    assert ch0_ch_mc.GetSize() == (128, 128)
    assert ch01_ch_mc.GetSize() == (128, 128, 2)
    assert ch12_ch_mc.GetSize() == (128, 128, 2)
    assert ch02_ch_mc.GetSize() == (128, 128, 2)
    assert ch1_ch_mc.GetSize() == (128, 128)
    assert ch2_ch_mc.GetSize() == (128, 128)
    assert std_gr.GetSize() == (128, 128)
    assert chsel0_gr.GetSize() == (128, 128)
    assert chsel01_gr.GetSize() == (128, 128)


@pytest.mark.usefixtures(
    "disk_im_gry_pyr",
    "disk_im_mch_pyr",
    "disk_im_rgb_pyr",
    "disk_im_gry",
    "disk_im_mch",
    "disk_im_rgb",
)
def test_tifffile_zarr_backend(
    disk_im_gry_pyr,
    disk_im_mch_pyr,
    disk_im_rgb_pyr,
    disk_im_gry,
    disk_im_mch,
    disk_im_rgb,
):
    im_gry_pyr = tifffile_zarr_backend(
        disk_im_gry_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_mch_pyr = tifffile_zarr_backend(
        disk_im_mch_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_rgb_pyr = tifffile_zarr_backend(
        disk_im_rgb_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="BF"),
    )
    im_gry = tifffile_zarr_backend(
        disk_im_gry,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_mch = tifffile_zarr_backend(
        disk_im_mch,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_rgb = tifffile_zarr_backend(
        disk_im_rgb,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="BF"),
    )

    assert im_gry_pyr.GetSize() == (2048, 2048)
    assert im_mch_pyr.GetSize() == (2048, 2048, 3)
    assert im_rgb_pyr.GetSize() == (2048, 2048)
    assert im_gry.GetSize() == (2048, 2048)
    assert im_mch.GetSize() == (2048, 2048, 3)
    assert im_rgb.GetSize() == (2048, 2048)


@pytest.mark.usefixtures(
    "disk_im_gry_pyr",
    "disk_im_mch_pyr",
    "disk_im_rgb_pyr",
    "disk_im_gry",
    "disk_im_mch",
    "disk_im_rgb",
)
def test_tifffile_dask_backend(
    disk_im_gry_pyr,
    disk_im_mch_pyr,
    disk_im_rgb_pyr,
    disk_im_gry,
    disk_im_mch,
    disk_im_rgb,
):
    im_gry_pyr = tifffile_dask_backend(
        disk_im_gry_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_mch_pyr = tifffile_dask_backend(
        disk_im_mch_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_rgb_pyr = tifffile_dask_backend(
        disk_im_rgb_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="BF"),
    )
    im_gry = tifffile_dask_backend(
        disk_im_gry,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_mch = tifffile_dask_backend(
        disk_im_mch,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_rgb = tifffile_dask_backend(
        disk_im_rgb,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="BF"),
    )

    assert im_gry_pyr.GetSize() == (2048, 2048)
    assert im_mch_pyr.GetSize() == (2048, 2048, 3)
    assert im_rgb_pyr.GetSize() == (2048, 2048)
    assert im_gry.GetSize() == (2048, 2048)
    assert im_mch.GetSize() == (2048, 2048, 3)
    assert im_rgb.GetSize() == (2048, 2048)


@pytest.mark.usefixtures(
    "disk_im_gry_pyr",
    "disk_im_mch_pyr",
    "disk_im_rgb_pyr",
    "disk_im_gry",
    "disk_im_mch",
    "disk_im_rgb",
)
def test_tifffile_dask_backend(
    disk_im_gry_pyr,
    disk_im_mch_pyr,
    disk_im_rgb_pyr,
    disk_im_gry,
    disk_im_mch,
    disk_im_rgb,
):
    im_gry_pyr = tifffile_dask_backend(
        disk_im_gry_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_mch_pyr = tifffile_dask_backend(
        disk_im_mch_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_rgb_pyr = tifffile_dask_backend(
        disk_im_rgb_pyr,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="BF"),
    )
    im_gry = tifffile_dask_backend(
        disk_im_gry,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_mch = tifffile_dask_backend(
        disk_im_mch,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="FL"),
    )
    im_rgb = tifffile_dask_backend(
        disk_im_rgb,
        largest_series=0,
        preprocessing=ImagePreproParams(image_type="BF"),
    )

    assert im_gry_pyr.GetSize() == (2048, 2048)
    assert im_mch_pyr.GetSize() == (2048, 2048, 3)
    assert im_rgb_pyr.GetSize() == (2048, 2048)
    assert im_gry.GetSize() == (2048, 2048)
    assert im_mch.GetSize() == (2048, 2048, 3)
    assert im_rgb.GetSize() == (2048, 2048)


@pytest.mark.usefixtures(
    "disk_im_gry",
    "disk_im_mch_notile",
    "disk_im_rgb",
)
def test_sitk_backend(
    disk_im_gry,
    disk_im_mch_notile,
    disk_im_rgb,
):

    im_gry = sitk_backend(
        str(disk_im_gry), preprocessing=ImagePreproParams(image_type="FL")
    )
    # im_mch = sitk_backend(str(disk_im_mch_notile), preprocessing={"image_type":"FL"})
    im_rgb = sitk_backend(
        str(disk_im_rgb), preprocessing=ImagePreproParams(image_type="BF")
    )

    assert im_gry.GetSize() == (2048, 2048)
    # assert im_mch.GetSize() == (2048,2048,3)
    assert im_rgb.GetSize() == (2048, 2048)


def test_guess_rgb():
    imshape_gs = (2048, 2048)
    imshape_mch = (3, 2048, 2048)
    imshape_rgb = (2048, 2048, 3)
    imshape_rgba = (2048, 2048, 4)

    assert guess_rgb(imshape_gs) is False
    assert guess_rgb(imshape_mch) is False
    assert guess_rgb(imshape_rgb) is True
    assert guess_rgb(imshape_rgba) is True


def test_greyscale():
    rgb_image_il = np.ones((2048, 2048, 3), dtype=np.uint8)
    rgb_image_noil = np.ones((3, 2048, 2048), dtype=np.uint8)

    da_rgb_image_il = da.ones(
        (2048, 2048, 3), dtype=np.uint8, chunks=(512, 512, 3)
    )
    da_rgb_image_noil = da.ones(
        (3, 2048, 2048), dtype=np.uint8, chunks=(512, 512, 3)
    )

    gs_il = grayscale(rgb_image_il, is_interleaved=True)
    gs_noil = grayscale(rgb_image_noil, is_interleaved=False)

    da_gs_il = grayscale(da_rgb_image_il, is_interleaved=True)
    da_gs_noil = grayscale(da_rgb_image_noil, is_interleaved=False)

    assert gs_il.shape == (2048, 2048)
    assert gs_noil.shape == (2048, 2048)
    assert np.array_equal(gs_il, gs_noil)
    assert da_gs_il.shape == (2048, 2048)
    assert da_gs_noil.shape == (2048, 2048)
    assert np.array_equal(da_gs_il.compute(), da_gs_noil.compute())


# PRIVATE_DIR = "tests/private_data"
@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_tile_greyscale():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.czi")
    czi = CziRegImageReader(image_fp)
    fsbd = czi.filtered_subblock_directory[5]
    subblock = fsbd.data_segment()
    tile = subblock.data(resize=False, order=0)
    tile_gs = czi_tile_grayscale(tile)

    assert tile.shape[:5] == tile_gs.shape[:5]
    assert tile_gs.shape[-1] == 1
    assert tile_gs.dtype == np.uint8


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_CziRegImageReader_rgb():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.czi")
    czi = CziRegImageReader(image_fp)
    gs_out = czi.sub_asarray_rgb(greyscale=True)
    rgb_out = czi.sub_asarray_rgb(greyscale=False)

    assert len(np.squeeze(gs_out).shape) == 2
    assert len(np.squeeze(rgb_out).shape) == 3


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_CziRegImageReader_mc():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.czi")
    czi = CziRegImageReader(image_fp)
    mc_out = czi.asarray(max_workers=1)
    ch0_out = czi.sub_asarray(channel_idx=[0], max_workers=1)
    ch1_out = czi.sub_asarray(channel_idx=[1], max_workers=1)
    ch2_out = czi.sub_asarray(channel_idx=[2], max_workers=1)
    ch3_out = czi.sub_asarray(channel_idx=[3], max_workers=1)
    ch02_out = czi.sub_asarray(channel_idx=[0, 2], max_workers=1)
    ch13_out = czi.sub_asarray(channel_idx=[1, 3], max_workers=1)
    ch03_out = czi.sub_asarray(channel_idx=[0, 3], max_workers=1)
    ch12_out = czi.sub_asarray(channel_idx=[1, 2], max_workers=1)
    ch23_out = czi.sub_asarray(channel_idx=[2, 3], max_workers=1)

    assert np.squeeze(mc_out).shape == (4, 4305, 4194)
    assert np.squeeze(ch0_out).shape == (4305, 4194)
    assert np.squeeze(ch1_out).shape == (4305, 4194)
    assert np.squeeze(ch1_out).shape == (4305, 4194)
    assert np.squeeze(ch2_out).shape == (4305, 4194)
    assert np.squeeze(ch3_out).shape == (4305, 4194)
    assert np.squeeze(ch02_out).shape == (2, 4305, 4194)
    assert np.array_equal(
        np.squeeze(ch02_out), np.squeeze(mc_out)[[0, 2], :, :]
    )
    assert np.array_equal(
        np.squeeze(ch13_out), np.squeeze(mc_out)[[1, 3], :, :]
    )
    assert np.array_equal(
        np.squeeze(ch03_out), np.squeeze(mc_out)[[0, 3], :, :]
    )
    assert np.array_equal(
        np.squeeze(ch12_out), np.squeeze(mc_out)[[1, 2], :, :]
    )
    assert np.array_equal(
        np.squeeze(ch23_out), np.squeeze(mc_out)[[2, 3], :, :]
    )
    assert np.array_equal(np.squeeze(mc_out)[0, :, :], np.squeeze(ch0_out))
    assert np.array_equal(np.squeeze(mc_out)[1, :, :], np.squeeze(ch1_out))
    assert np.array_equal(np.squeeze(mc_out)[2, :, :], np.squeeze(ch2_out))
    assert np.array_equal(np.squeeze(mc_out)[3, :, :], np.squeeze(ch3_out))


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_tf_get_largest_series():
    image_fp_czi_mc = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.ome.tiff")
    image_fp_czi_rgb = os.path.join(PRIVATE_DIR, "czi_rgb.ome.tiff")
    image_fp_bigtiff = os.path.join(PRIVATE_DIR, "huron_rgb.tif")
    image_fp_scn = os.path.join(PRIVATE_DIR, "scn_rgb.scn")
    assert tf_get_largest_series(image_fp_czi_mc) == 0
    assert tf_get_largest_series(image_fp_czi_rgb) == 0
    assert tf_get_largest_series(image_fp_bigtiff) == 0
    assert tf_get_largest_series(image_fp_scn) == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_get_sitk_image_info():
    image_fp_jpgrgb = os.path.join(PRIVATE_DIR, "testjpegrgb.jpg")
    image_fp_pngrgb = os.path.join(PRIVATE_DIR, "testpngrgb.png")

    assert np.array_equal(
        get_sitk_image_info(image_fp_jpgrgb)[0], [450, 750, 3]
    )
    assert get_sitk_image_info(image_fp_jpgrgb)[1] == np.uint8
    assert np.array_equal(
        get_sitk_image_info(image_fp_pngrgb)[0], [450, 750, 3]
    )
    assert get_sitk_image_info(image_fp_pngrgb)[1] == np.uint8


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_get_tifffile_info_private():
    image_fp_bigtiff = os.path.join(PRIVATE_DIR, "huron_rgb.tif")
    image_fp_scn = os.path.join(PRIVATE_DIR, "scn_rgb.scn")

    assert np.array_equal(
        get_tifffile_info(image_fp_bigtiff)[0], [7662, 15778, 3]
    )
    assert get_tifffile_info(image_fp_bigtiff)[1] == np.uint8
    assert np.array_equal(
        get_tifffile_info(image_fp_scn)[0], [11776, 18528, 3]
    )
    assert get_tifffile_info(image_fp_scn)[1] == np.uint8


@pytest.mark.usefixtures(
    "disk_im_gry_pyr",
    "disk_im_mch_pyr",
    "disk_im_rgb_pyr",
)
def test_get_tifffile_info_public(
    disk_im_gry_pyr,
    disk_im_mch_pyr,
    disk_im_rgb_pyr,
):

    assert np.array_equal(
        get_tifffile_info(disk_im_rgb_pyr)[0], [2048, 2048, 3]
    )
    assert get_tifffile_info(disk_im_rgb_pyr)[1] == np.uint8

    assert np.array_equal(
        get_tifffile_info(disk_im_mch_pyr)[0], [3, 2048, 2048]
    )
    assert get_tifffile_info(disk_im_mch_pyr)[1] == np.uint16

    assert np.array_equal(
        get_tifffile_info(disk_im_gry_pyr)[0], [1, 2048, 2048]
    )
    assert get_tifffile_info(disk_im_gry_pyr)[1] == np.uint16



import yaml


def parse_check_reg_config(yaml_filepath):
    with open(yaml_filepath, "r") as file:
        reg_config = yaml.full_load(file)

    def check_for_key(top_key, check_dict, check_key):
        if check_dict.get(check_key) is None:
            raise ValueError(f"{top_key} does not contain an {check_key}")

    if reg_config.get("project_name") is None:
        raise ValueError(
            f"{yaml_filepath} does not contain a project_name key"
        )

    if reg_config.get("output_dir") is None:
        raise ValueError(f"{yaml_filepath} does not contain a output_dir key")

    if reg_config.get("cache_images") is None:
        reg_config.update({"cache_images": True})

    if reg_config.get("modalities"):
        for key, val in reg_config["modalities"].items():
            [
                check_for_key(key, val, ck)
                for ck in ["image_filepath", "image_res"]
            ]

    if reg_config.get("reg_paths"):

        for key, val in reg_config["reg_paths"].items():
            [
                check_for_key(key, val, ck)
                for ck in [
                    "src_modality_name",
                    "tgt_modality_name",
                    "reg_params",
                ]
            ]
            if isinstance(val.get("reg_params"), str):
                val.update({"reg_params": [val.get("reg_params")]})

    if reg_config.get("attachment_images"):

        for key, val in reg_config["attachment_images"].items():
            [
                check_for_key(key, val, ck)
                for ck in [
                    "attachment_modality",
                    "image_filepath",
                    "image_res",
                ]
            ]

    if reg_config.get("attachment_shapes"):

        for key, val in reg_config["attachment_shapes"].items():
            [
                check_for_key(key, val, ck)
                for ck in ["attachment_modality", "shape_files"]
            ]

    return reg_config



import numpy as np
import SimpleITK as sitk

from wsireg.reg_images.reg_image import RegImage
from wsireg.utils.im_utils import (
    ensure_dask_array,
    get_sitk_image_info,
    guess_rgb,
    sitk_vect_to_gs,
)


class SitkRegImage(RegImage):
    def __init__(
        self,
        image,
        image_res,
        mask=None,
        pre_reg_transforms=None,
        preprocessing=None,
        channel_names=None,
        channel_colors=None,
    ):
        super(SitkRegImage, self).__init__(preprocessing)
        self.image_filepath = image
        self.image_res = image_res
        self.image = None

        self.reader = "sitk"

        (
            self.im_dims,
            self.im_dtype,
        ) = self._get_image_info()

        self.im_dims = tuple(self.im_dims)
        self.is_rgb = guess_rgb(self.im_dims)

        self.n_ch = self.im_dims[2] if self.is_rgb else self.im_dims[0]

        self.reg_image = None
        self.mask = self.read_mask(mask)

        self.pre_reg_transforms = pre_reg_transforms

        self.channel_names = channel_names
        self.channel_colors = channel_colors
        self.original_size_transform = None

    def _get_image_info(self):

        im_dims, im_dtype = get_sitk_image_info(self.image_filepath)

        return im_dims, im_dtype

    def read_reg_image(self):
        reg_image = sitk.ReadImage(self.image_filepath)

        if reg_image.GetNumberOfComponentsPerPixel() >= 3:
            if self.preprocessing is not None:
                reg_image = sitk_vect_to_gs(reg_image)

        if (
            self.preprocessing.get("as_uint8") is True
            and reg_image.GetPixelID() != 1
        ):
            reg_image = sitk.Cast(
                sitk.RescaleIntensity(reg_image), sitk.sitkUInt8
            )

        if (
            self.preprocessing.get("ch_indices") is not None
            and reg_image.GetDepth() > 0
        ):
            chs = np.asarray(self.preprocessing.get('ch_indices'))
            reg_image = reg_image[:, :, chs]

        self.preprocess_image(reg_image)

    def read_full_image(self):
        self.image = ensure_dask_array(
            sitk.GetArrayFromImage(sitk.ReadImage(self.image_filepath))
        )



import numpy as np
import SimpleITK as sitk

from wsireg.reg_images import RegImage
from wsireg.utils.im_utils import CziRegImageReader, guess_rgb


class CziRegImage(RegImage):
    def __init__(
        self,
        image,
        image_res,
        mask=None,
        pre_reg_transforms=None,
        preprocessing=None,
        channel_names=None,
        channel_colors=None,
    ):
        super(CziRegImage, self).__init__(preprocessing)
        self.image_filepath = image
        self.image_res = image_res
        self.image = None
        self.czi = CziRegImageReader(self.image_filepath)
        self.reader = "czi"

        (
            self.ch_dim_idx,
            self.y_dim_idx,
            self.x_dim_idx,
            self.im_dims,
            self.im_dtype,
        ) = self._get_image_info()

        self.im_dims = tuple(self.im_dims)
        self.is_rgb = guess_rgb(self.im_dims)
        self.n_ch = self.im_dims[2] if self.is_rgb else self.im_dims[0]

        self.reg_image = None
        self.mask = self.read_mask(mask)

        self.pre_reg_transforms = pre_reg_transforms

        self.channel_names = channel_names
        self.channel_colors = channel_colors
        self.original_size_transform = None

    def _get_image_info(self):
        # if RGB need to get 0
        if self.czi.shape[-1] > 1:
            ch_dim_idx = self.czi.axes.index('0')
        else:
            ch_dim_idx = self.czi.axes.index('C')
        y_dim_idx = self.czi.axes.index('Y')
        x_dim_idx = self.czi.axes.index('X')
        if self.czi.shape[-1] > 1:
            im_dims = np.array(self.czi.shape)[
                [y_dim_idx, x_dim_idx, ch_dim_idx]
            ]
        else:
            im_dims = np.array(self.czi.shape)[
                [ch_dim_idx, y_dim_idx, x_dim_idx]
            ]

        im_dtype = self.czi.dtype

        return ch_dim_idx, y_dim_idx, x_dim_idx, im_dims, im_dtype

    def read_reg_image(self):
        scene_idx = self.czi.axes.index('S')

        if self.czi.shape[scene_idx] > 1:
            raise ValueError('multi scene czis not allowed at this time')
        if self.is_rgb:
            reg_image = self.czi.sub_asarray_rgb(greyscale=True)
        else:
            reg_image = self.czi.sub_asarray(
                channel_idx=self.preprocessing.ch_indices,
                as_uint8=self.preprocessing.as_uint8,
            )

        reg_image = np.squeeze(reg_image)
        reg_image = sitk.GetImageFromArray(reg_image)

        self.preprocess_image(reg_image)

    def read_single_channel(self, channel_idx: int):
        if self.is_rgb is False:
            image = self.czi.sub_asarray(
                channel_idx=[channel_idx],
            )
        else:
            image = self.czi.sub_asarray_rgb(
                channel_idx=[channel_idx], greyscale=False
            )

        return image



import itk
import pytest
import SimpleITK as sitk

from wsireg.reg_images.loader import reg_image_loader


@pytest.mark.usefixtures("disk_im_mch")
def test_reg_image_loader_image_fp_mc_std_prepro(disk_im_mch):
    reg_image = reg_image_loader(str(disk_im_mch), 0.65)
    reg_image.read_reg_image()
    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("disk_im_rgb")
def test_reg_image_loader_image_fp_rgb_std_prepro(disk_im_rgb):
    reg_image = reg_image_loader(str(disk_im_rgb), 0.65)
    reg_image.read_reg_image()
    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("disk_im_gry")
def test_reg_image_loader_image_fp_gry_std_prepro(disk_im_gry):
    reg_image = reg_image_loader(str(disk_im_gry), 0.65)
    reg_image.read_reg_image()
    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_mch_np")
def test_reg_image_loader_image_np_mc_std_prepro(im_mch_np):
    reg_image = reg_image_loader(im_mch_np, 0.65)
    reg_image.read_reg_image()
    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetDepth() == 0


@pytest.mark.usefixtures("im_rgb_np")
def test_reg_image_loader_image_np_rgb_std_prepro(im_rgb_np):
    reg_image = reg_image_loader(im_rgb_np, 0.65)
    reg_image.read_reg_image()
    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.usefixtures("im_gry_np")
def test_reg_image_loader_image_np_gry_std_prepro(im_gry_np):
    reg_image = reg_image_loader(im_gry_np, 0.65)
    reg_image.read_reg_image()
    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.usefixtures("im_mch_np")
def test_reg_image_loader_image_np_mc_std_prepro_rot(im_mch_np):
    reg_image = reg_image_loader(im_mch_np, 0.65, preprocessing={"rot_cc": 90})
    reg_image.read_reg_image()
    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetDepth() == 0
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_rgb_np")
def test_reg_image_loader_image_np_rgb_std_prepro_rot(im_rgb_np):
    reg_image = reg_image_loader(im_rgb_np, 0.65, preprocessing={"rot_cc": 90})
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np")
def test_reg_image_loader_image_np_gry_std_prepro_rot(im_gry_np):
    reg_image = reg_image_loader(im_gry_np, 0.65, preprocessing={"rot_cc": 90})
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_mch_np")
def test_reg_image_loader_image_np_mc_std_prepro_fliph(im_mch_np):
    reg_image = reg_image_loader(im_mch_np, 0.65, preprocessing={"flip": "h"})
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetDepth() == 0
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_rgb_np")
def test_reg_image_loader_image_np_rgb_std_prepro_fliph(im_rgb_np):
    reg_image = reg_image_loader(im_rgb_np, 0.65, preprocessing={"flip": "h"})
    reg_image.read_reg_image()
    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np")
def test_reg_image_loader_image_np_gry_std_prepro_fliph(im_gry_np):
    reg_image = reg_image_loader(im_gry_np, 0.65, preprocessing={"flip": "h"})
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1


@pytest.mark.usefixtures("im_mch_np")
def test_reg_image_loader_image_np_mc_std_prepro_flipv(im_mch_np):
    reg_image = reg_image_loader(im_mch_np, 0.65, preprocessing={"flip": "v"})
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetDepth() == 0
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_rgb_np")
def test_reg_image_loader_image_np_rgb_std_prepro_flipv(im_rgb_np):
    reg_image = reg_image_loader(im_rgb_np, 0.65, preprocessing={"flip": "v"})
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np")
def test_reg_image_loader_image_np_gry_std_prepro_flipv(im_gry_np):
    reg_image = reg_image_loader(im_gry_np, 0.65, preprocessing={"flip": "v"})
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_mch_np")
def test_reg_image_loader_image_np_mc_std_prepro_rot_flipv(im_mch_np):
    reg_image = reg_image_loader(
        im_mch_np, 0.65, preprocessing={"rot_cc": 90, "flip": "v"}
    )
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.reg_image.GetDepth() == 0
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 2
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np")
def test_reg_image_loader_image_np_gry_std_prepro_flipv(im_gry_np):
    reg_image = reg_image_loader(im_gry_np, 0.65, preprocessing={"flip": "v"})
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (2048, 2048)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert reg_image.pre_reg_transforms is not None
    assert len(reg_image.pre_reg_transforms) == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np", "mask_np")
def test_reg_image_loader_mask(im_gry_np, mask_np):
    reg_image = reg_image_loader(im_gry_np, 0.65, mask=mask_np)
    reg_image.read_reg_image()
    assert reg_image.mask is not None
    assert isinstance(reg_image.mask, sitk.Image) is True
    assert reg_image.mask.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np", "mask_np")
def test_reg_image_loader_mask_rot(im_gry_np, mask_np):
    reg_image = reg_image_loader(
        im_gry_np, 0.65, preprocessing={"rot_cc": 90}, mask=mask_np
    )
    reg_image.read_reg_image()

    assert reg_image.mask is not None
    assert isinstance(reg_image.mask, sitk.Image) is True
    assert reg_image.mask.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np", "mask_np")
def test_reg_image_loader_mask_rot_flip(im_gry_np, mask_np):
    reg_image = reg_image_loader(
        im_gry_np,
        0.65,
        preprocessing={"rot_cc": 90, "flip": "h"},
        mask=mask_np,
    )
    reg_image.read_reg_image()

    assert reg_image.mask is not None
    assert isinstance(reg_image.mask, sitk.Image) is True
    assert reg_image.mask.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np", "mask_np")
def test_reg_image_loader_mask_flip(im_gry_np, mask_np):
    reg_image = reg_image_loader(
        im_gry_np, 0.65, preprocessing={"flip": "v"}, mask=mask_np
    )
    reg_image.read_reg_image()

    assert reg_image.mask is not None
    assert isinstance(reg_image.mask, sitk.Image) is True
    assert reg_image.mask.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np", "mask_np")
def test_reg_image_loader_downsampling(im_gry_np, mask_np):
    reg_image = reg_image_loader(
        im_gry_np, 1, preprocessing={"downsampling": 2}, mask=mask_np
    )
    reg_image.read_reg_image()

    assert reg_image.reg_image.GetSize() == (1024, 1024)
    assert reg_image.mask.GetSize() == (1024, 1024)
    assert reg_image.mask.GetSpacing() == (2, 2)
    assert reg_image.reg_image.GetSpacing() == (2, 2)


@pytest.mark.usefixtures("im_gry_np", "mask_np")
def test_reg_image_loader_to_itk(im_gry_np, mask_np):
    reg_image = reg_image_loader(im_gry_np, 0.65, mask=mask_np)
    reg_image.read_reg_image()
    reg_image.reg_image_sitk_to_itk(cast_to_float32=True)
    assert isinstance(reg_image.reg_image, itk.Image) is True
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert isinstance(reg_image.mask, itk.Image) is True
    assert reg_image.mask.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("im_gry_np", "mask_np")
def test_reg_image_loader_to_itk(im_gry_np, mask_np):
    reg_image = reg_image_loader(im_gry_np, 0.65, mask=mask_np)
    reg_image.read_reg_image()
    reg_image.reg_image_sitk_to_itk(cast_to_float32=True)
    assert isinstance(reg_image.reg_image, itk.Image) is True
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert isinstance(reg_image.mask, itk.Image) is True
    assert reg_image.mask.GetSpacing() == (0.65, 0.65)


@pytest.mark.usefixtures("dask_im_rgb_np")
def test_reg_image_loader_dask_rgb(dask_im_rgb_np):
    reg_image = reg_image_loader(dask_im_rgb_np, 0.65)
    reg_image.read_reg_image()
    assert len(reg_image.im_dims) == 3
    assert reg_image.im_dims[-1] == 3
    assert reg_image.is_rgb
    assert reg_image.n_ch == 3
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.usefixtures("dask_im_gry_np")
def test_reg_image_loader_dask_gry(dask_im_gry_np):
    reg_image = reg_image_loader(dask_im_gry_np, 0.65)
    reg_image.read_reg_image()
    assert len(reg_image.im_dims) == 3
    assert reg_image.im_dims[0] == 1
    assert reg_image.is_rgb is False
    assert reg_image.n_ch == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.usefixtures("dask_im_mch_np")
def test_reg_image_loader_dask_mch(dask_im_mch_np):
    reg_image = reg_image_loader(dask_im_mch_np, 0.65)
    reg_image.read_reg_image()
    assert len(reg_image.im_dims) == 3
    assert reg_image.im_dims[0] == 3
    assert reg_image.is_rgb is False
    assert reg_image.n_ch == 3
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.usefixtures("zarr_im_rgb_np")
def test_reg_image_loader_zarr_rgb(zarr_im_rgb_np):
    reg_image = reg_image_loader(zarr_im_rgb_np, 0.65)
    reg_image.read_reg_image()
    assert len(reg_image.im_dims) == 3
    assert reg_image.im_dims[-1] == 3
    assert reg_image.is_rgb
    assert reg_image.n_ch == 3
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.usefixtures("zarr_im_gry_np")
def test_reg_image_loader_zarr_gry(zarr_im_gry_np):
    reg_image = reg_image_loader(zarr_im_gry_np, 0.65)
    reg_image.read_reg_image()
    assert len(reg_image.im_dims) == 3
    assert reg_image.im_dims[0] == 1
    assert reg_image.is_rgb is False
    assert reg_image.n_ch == 1
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.usefixtures("zarr_im_mch_np")
def test_reg_image_loader_zarr_mch(zarr_im_mch_np):
    reg_image = reg_image_loader(zarr_im_mch_np, 0.65)
    reg_image.read_reg_image()
    assert len(reg_image.im_dims) == 3
    assert reg_image.im_dims[0] == 3
    assert reg_image.is_rgb is False
    assert reg_image.n_ch == 3
    assert reg_image.reg_image.GetSpacing() == (0.65, 0.65)
    assert reg_image.reg_image.GetNumberOfComponentsPerPixel() == 1



import multiprocessing
import warnings
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import cv2
import dask.array as da
import numpy as np
import SimpleITK as sitk
import zarr
from czifile import CziFile
from tifffile import (
    OmeXml,
    TiffFile,
    TiffWriter,
    create_output,
    imread,
    xml2dict,
)

from wsireg.parameter_maps.preprocessing import BoundingBox
from wsireg.utils.tform_utils import sitk_transform_image

TIFFFILE_EXTS = [".scn", ".tif", ".tiff", ".ndpi", ".svs"]

ARRAYLIKE_CLASSES = (np.ndarray, da.core.Array, zarr.Array)

SITK_TO_NP_DTYPE = {
    0: np.int8,
    1: np.uint8,
    2: np.int16,
    3: np.uint16,
    4: np.int32,
    5: np.uint32,
    6: np.int64,
    7: np.uint64,
    8: np.float32,
    9: np.float64,
    10: np.complex64,
    11: np.complex64,
    12: np.int8,
    13: np.uint8,
    14: np.int16,
    15: np.int16,
    16: np.int32,
    17: np.int32,
    18: np.int64,
    19: np.int64,
    20: np.float32,
    21: np.float64,
    22: np.uint8,
    23: np.uint16,
    24: np.uint32,
    25: np.uint64,
}

COLNAME_TO_HEX = {
    "red": "FF0000",
    "green": "00FF00",
    "blue": "0000FF",
    "magenta": "FF00FF",
    "yellow": "FFFF00",
    "cyan": "00FFFFF",
    "white": "FFFFFF",
}


def zarr_get_base_pyr_layer(zarr_store):
    """
    Find the base pyramid layer of a zarr store

    Parameters
    ----------
    zarr_store
        zarr store

    Returns
    -------
    zarr_im: zarr.core.Array
        zarr array of base layer
    """
    if isinstance(zarr_store, zarr.hierarchy.Group):
        zarr_im = zarr_store[str(0)]
    elif isinstance(zarr_store, zarr.core.Array):
        zarr_im = zarr_store
    return zarr_im


def ensure_dask_array(image):
    if isinstance(image, da.core.Array):
        return image

    if isinstance(image, zarr.Array):
        return da.from_zarr(image)

    # handles np.ndarray _and_ other array like objects.
    return da.from_array(image)


def read_preprocess_array(array, preprocessing, force_rgb=None):
    """Read np.array, zarr.Array, or dask.array image into memory
    with preprocessing for registration."""
    is_interleaved = guess_rgb(array.shape)
    is_rgb = is_interleaved if not force_rgb else force_rgb

    if is_rgb:
        if preprocessing:
            image_out = np.asarray(
                grayscale(array, is_interleaved=is_interleaved)
            )
            image_out = sitk.GetImageFromArray(image_out)
        else:
            image_out = np.asarray(array)
            if not is_interleaved:
                image_out = np.rollaxis(image_out, 0, 3)
            image_out = sitk.GetImageFromArray(image_out, isVector=True)

    elif len(array.shape) == 2:
        image_out = sitk.GetImageFromArray(np.asarray(array))

    else:
        if preprocessing:
            if preprocessing.ch_indices and len(array.shape) > 2:
                chs = list(preprocessing.ch_indices)
                array = array[chs, :, :]

        image_out = sitk.GetImageFromArray(np.squeeze(np.asarray(array)))

    return image_out


def tifffile_zarr_backend(
    image_filepath, largest_series, preprocessing, force_rgb=None
):
    """
    Read image with tifffile and use zarr to read data into memory

    Parameters
    ----------
    image_filepath: str
        path to the image file
    largest_series: int
        index of the largest series in the image
    preprocessing:
        whether to do some read-time pre-processing
        - greyscale conversion (at the tile level)
        - read individual or range of channels (at the tile level)

    Returns
    -------
    image: sitk.Image
        image ready for other registration pre-processing

    """
    print("using zarr backend")
    zarr_series = imread(image_filepath, aszarr=True, series=largest_series)
    zarr_store = zarr.open(zarr_series)
    zarr_im = zarr_get_base_pyr_layer(zarr_store)
    return read_preprocess_array(
        zarr_im, preprocessing=preprocessing, force_rgb=force_rgb
    )


def tifffile_dask_backend(
    image_filepath, largest_series, preprocessing, force_rgb=None
):
    """
    Read image with tifffile and use dask to read data into memory

    Parameters
    ----------
    image_filepath: str
        path to the image file
    largest_series: int
        index of the largest series in the image
    preprocessing:
        whether to do some read-time pre-processing
        - greyscale conversion (at the tile level)
        - read individual or range of channels (at the tile level)

    Returns
    -------
    image: sitk.Image
        image ready for other registration pre-processing

    """
    print("using dask backend")
    zarr_series = imread(image_filepath, aszarr=True, series=largest_series)
    zarr_store = zarr.open(zarr_series)
    dask_im = da.squeeze(da.from_zarr(zarr_get_base_pyr_layer(zarr_store)))
    return read_preprocess_array(
        dask_im, preprocessing=preprocessing, force_rgb=force_rgb
    )


def sitk_backend(image_filepath, preprocessing):
    """
    Read image with SimpleITK..this will always read the full image into memory

    Parameters
    ----------
    image_filepath: str
        path to the image file
    preprocessing:
        whether to do some read-time pre-processing
        - greyscale conversion (at the tile level)
        - read individual or range of channels (at the tile level)

    Returns
    -------
    image: sitk.Image
        image ready for other registration pre-processing

    """
    print("using sitk backend")
    image = sitk.ReadImage(image_filepath)

    if image.GetNumberOfComponentsPerPixel() >= 3:
        if preprocessing is not None:
            image = sitk_vect_to_gs(image)

    elif image.GetDepth() == 0:
        return image
    else:
        if preprocessing is not None:
            if (
                preprocessing.get("ch_indices") is not None
                and image.GetDepth() > 0
            ):
                print("here")
                chs = np.asarray(preprocessing.get('ch_indices'))
                image = image[:, :, chs]

    return image


def guess_rgb(shape):
    """
    Guess if the passed shape comes from rgb data.
    If last dim is 3 or 4 assume the data is rgb, including rgba.

    Parameters
    ----------
    shape : list of int
        Shape of the data that should be checked.

    Returns
    -------
    bool
        If data is rgb or not.
    """
    ndim = len(shape)
    last_dim = shape[-1]
    if ndim > 2 and last_dim < 5:
        rgb = True
    else:
        rgb = False

    return rgb


def grayscale(rgb_image, is_interleaved=False):
    """
    convert RGB image data to greyscale

    Parameters
    ----------
    rgb_image: np.ndarray
        image data
    Returns
    -------
    image:np.ndarray
        returns 8-bit greyscale image for 24-bit RGB image
    """
    if is_interleaved is True:
        result = (
            (rgb_image[..., 0] * 0.2125).astype(np.uint8)
            + (rgb_image[..., 1] * 0.7154).astype(np.uint8)
            + (rgb_image[..., 2] * 0.0721).astype(np.uint8)
        )
    else:
        result = (
            (rgb_image[0, ...] * 0.2125).astype(np.uint8)
            + (rgb_image[1, ...] * 0.7154).astype(np.uint8)
            + (rgb_image[2, ...] * 0.0721).astype(np.uint8)
        )

    return result


def czi_tile_grayscale(rgb_image):
    """
    convert RGB image data to greyscale

    Parameters
    ----------
    rgb_image: np.ndarray
        image data
    Returns
    -------
    image:np.ndarray
        returns 8-bit greyscale image for 24-bit RGB image
    """
    result = (
        (rgb_image[..., 0] * 0.2125).astype(np.uint8)
        + (rgb_image[..., 1] * 0.7154).astype(np.uint8)
        + (rgb_image[..., 2] * 0.0721).astype(np.uint8)
    )

    return np.expand_dims(result, axis=-1)


class CziRegImageReader(CziFile):
    """
    Sub-class of CziFile with added functionality to only read certain channels
    """

    def sub_asarray(
        self,
        resize=True,
        order=0,
        out=None,
        max_workers=None,
        channel_idx=None,
        as_uint8=False,
    ):

        """Return image data from file(s) as numpy array.

        Parameters
        ----------
        resize : bool
            If True (default), resize sub/supersampled subblock data.
        order : int
            The order of spline interpolation used to resize sub/supersampled
            subblock data. Default is 0 (nearest neighbor).
        out : numpy.ndarray, str, or file-like object; optional
            Buffer where image data will be saved.
            If numpy.ndarray, a writable array of compatible dtype and shape.
            If str or open file, the file name or file object used to
            create a memory-map to an array stored in a binary file on disk.
        max_workers : int
            Maximum number of threads to read and decode subblock data.
            By default up to half the CPU cores are used.
        channel_idx : int or list of int
            The indices of the channels to extract
        as_uint8 : bool
            byte-scale image data to np.uint8 data type

        Parameters
        ----------
        out:np.ndarray
            image read with selected parameters as np.ndarray
        """

        out_shape = list(self.shape)
        start = list(self.start)

        ch_dim_idx = self.axes.index('C')

        if channel_idx is not None:
            if isinstance(channel_idx, int):
                channel_idx = [channel_idx]

            if out_shape[ch_dim_idx] == 1:
                channel_idx = None

            else:
                out_shape[ch_dim_idx] = len(channel_idx)
                min_ch_seq = {}
                for idx, i in enumerate(channel_idx):
                    min_ch_seq.update({i: idx})

        if as_uint8 is True:
            out_dtype = np.uint8
        else:
            out_dtype = self.dtype

        if out is None:
            out = create_output(None, tuple(out_shape), out_dtype)

        if max_workers is None:
            max_workers = multiprocessing.cpu_count() - 1

        def func(
            directory_entry, resize=resize, order=order, start=start, out=out
        ):
            """Read, decode, and copy subblock data."""
            subblock = directory_entry.data_segment()
            dvstart = list(directory_entry.start)
            czi_c_idx = [
                de.dimension for de in subblock.dimension_entries
            ].index('C')
            subblock_ch_idx = subblock.dimension_entries[czi_c_idx].start
            if channel_idx is not None:
                if subblock_ch_idx in channel_idx:
                    subblock.dimension_entries[czi_c_idx].start
                    tile = subblock.data(resize=resize, order=order)
                    dvstart[ch_dim_idx] = min_ch_seq.get(subblock_ch_idx)
                else:
                    return
            else:
                tile = subblock.data(resize=resize, order=order)

            if as_uint8 is True:
                tile = (tile / 256).astype("uint8")

            index = tuple(
                slice(i - j, i - j + k)
                for i, j, k in zip(tuple(dvstart), tuple(start), tile.shape)
            )

            try:
                out[index] = tile
            except ValueError as e:
                warnings.warn(str(e))

        if max_workers > 1:
            self._fh.lock = True
            with ThreadPoolExecutor(max_workers) as executor:
                executor.map(func, self.filtered_subblock_directory)
            self._fh.lock = None
        else:
            for directory_entry in self.filtered_subblock_directory:
                func(directory_entry)

        if hasattr(out, "flush"):
            out.flush()
        return out

    def sub_asarray_rgb(
        self,
        resize=True,
        order=0,
        out=None,
        max_workers=None,
        channel_idx=None,
        as_uint8=False,
        greyscale=False,
    ):

        """Return image data from file(s) as numpy array.

        Parameters
        ----------
        resize : bool
            If True (default), resize sub/supersampled subblock data.
        order : int
            The order of spline interpolation used to resize sub/supersampled
            subblock data. Default is 0 (nearest neighbor).
        out : numpy.ndarray, str, or file-like object; optional
            Buffer where image data will be saved.
            If numpy.ndarray, a writable array of compatible dtype and shape.
            If str or open file, the file name or file object used to
            create a memory-map to an array stored in a binary file on disk.
        max_workers : int
            Maximum number of threads to read and decode subblock data.
            By default up to half the CPU cores are used.
        channel_idx : int or list of int
            The indices of the channels to extract
        as_uint8 : bool
            byte-scale image data to np.uint8 data type

        Parameters
        ----------
        out:np.ndarray
            image read with selected parameters as np.ndarray
        """

        out_shape = list(self.shape)
        start = list(self.start)
        ch_dim_idx = self.axes.index('0')

        if channel_idx is not None:
            if isinstance(channel_idx, int):
                channel_idx = [channel_idx]
            out_shape[ch_dim_idx] = len(channel_idx)

        if greyscale is True:
            out_shape[ch_dim_idx] = 1

        if as_uint8 is True:
            out_dtype = np.uint8
        else:
            out_dtype = self.dtype

        if out is None:
            out = create_output(None, tuple(out_shape), out_dtype)

        if max_workers is None:
            max_workers = multiprocessing.cpu_count() - 1

        def func(
            directory_entry, resize=resize, order=order, start=start, out=out
        ):
            """Read, decode, and copy subblock data."""
            subblock = directory_entry.data_segment()
            dvstart = list(directory_entry.start)
            tile = subblock.data(resize=resize, order=order)

            if greyscale is True:
                tile = czi_tile_grayscale(tile)

            if channel_idx is not None:
                tile = tile[:, :, :, :, :, channel_idx]

            index = tuple(
                slice(i - j, i - j + k)
                for i, j, k in zip(tuple(dvstart), tuple(start), tile.shape)
            )

            try:
                out[index] = tile
            except ValueError as e:
                warnings.warn(str(e))

        if max_workers > 1:
            self._fh.lock = True
            with ThreadPoolExecutor(max_workers) as executor:
                executor.map(func, self.filtered_subblock_directory)
            self._fh.lock = None
        else:
            for directory_entry in self.filtered_subblock_directory:
                func(directory_entry)

        if hasattr(out, "flush"):
            out.flush()
        return out


def tf_get_largest_series(image_filepath):
    """
    Determine largest series for .scn files by examining metadata
    For other multi-series files, find the one with the most pixels

    Parameters
    ----------
    image_filepath: str
        path to the image file

    Returns
    -------
    largest_series:int
        index of the largest series in the image data
    """
    fp_ext = Path(image_filepath).suffix.lower()
    tf_im = TiffFile(image_filepath)
    if fp_ext == ".scn":
        scn_meta = xml2dict(tf_im.scn_metadata)
        image_meta = scn_meta.get("scn").get("collection").get("image")
        largest_series = np.argmax(
            [
                im.get("scanSettings")
                .get("objectiveSettings")
                .get("objective")
                for im in image_meta
            ]
        )
    else:
        largest_series = np.argmax(
            [
                np.prod(np.asarray(series.shape), dtype=np.int64)
                for series in tf_im.series
            ]
        )
    return largest_series


def get_sitk_image_info(image_filepath):
    """
    Get image info for files only ready by SimpleITK

    Parameters
    ----------
    image_filepath:str
        filepath to image

    Returns
    -------
    im_dims: np.ndarray
        image dimensions in np.ndarray
    im_dtype: np.dtype
        data type of the image

    """
    reader = sitk.ImageFileReader()
    reader.SetFileName(image_filepath)
    reader.LoadPrivateTagsOn()
    reader.ReadImageInformation()
    im_dims = np.asarray(reader.GetSize())
    # swap to YX
    im_dims[[0, 1]] = im_dims[[1, 0]]
    im_dtype = np.dtype(SITK_TO_NP_DTYPE.get(reader.GetPixelID()))
    is_vector = sitk.GetPixelIDValueAsString(reader.GetPixelID())

    if "vector" in is_vector:
        im_dims = np.append(im_dims, 3)
    elif len(im_dims) == 3:
        im_dims = im_dims[[2, 0, 1]]
    else:
        im_dims = np.concatenate([[1], im_dims], axis=0)

    return im_dims, im_dtype


def get_tifffile_info(image_filepath):
    largest_series = tf_get_largest_series(image_filepath)
    zarr_im = zarr.open(
        imread(image_filepath, aszarr=True, series=largest_series)
    )
    zarr_im = zarr_get_base_pyr_layer(zarr_im)
    im_dims = np.squeeze(zarr_im.shape)
    if len(im_dims) == 2:
        im_dims = np.concatenate([[1], im_dims])
    im_dtype = zarr_im.dtype

    return im_dims, im_dtype


def tf_zarr_read_single_ch(
    image_filepath, channel_idx, is_rgb, is_rgb_interleaved=True
):
    """
    Reads a single channel using zarr or dask in combination with tifffile

    Parameters
    ----------
    image_filepath:str
        file path to image
    channel_idx:int
        index of the channel to be read
    is_rgb:bool
        whether image is rgb interleaved

    Returns
    -------
    im:np.ndarray
        image as a np.ndarray
    """
    largest_series = tf_get_largest_series(image_filepath)
    zarr_im = zarr.open(
        imread(image_filepath, aszarr=True, series=largest_series)
    )
    zarr_im = zarr_get_base_pyr_layer(zarr_im)
    try:
        im = da.squeeze(da.from_zarr(zarr_im))
        if is_rgb and is_rgb_interleaved is True:
            im = im[:, :, channel_idx].compute()
        elif len(im.shape) > 2:
            im = im[channel_idx, :, :].compute()
        else:
            im = im.compute()

    except ValueError:
        im = zarr_im
        if is_rgb is True and is_rgb_interleaved is True:
            im = im[:, :, channel_idx]
        elif len(im.shape) > 2:
            im = im[channel_idx, :, :].compute()
        else:
            im = im.compute()
    return im


def czi_read_single_ch(image_filepath, channel_idx):
    czi = CziRegImageReader(image_filepath)

    im = czi.sub_asarray(
        channel_idx=channel_idx,
    )

    return im


def calc_pyramid_levels(xy_final_shape, tile_size):
    """
    Calculate number of pyramids for a given image dimension and tile size
    Stops when further downsampling would be smaller than tile_size.

    Parameters
    ----------
    xy_final_shape:np.ndarray
        final shape in xy order
    tile_size: int
        size of the tiles in the pyramidal layers
    Returns
    -------
    res_shapes:list
        list of tuples of the shapes of the downsampled images

    """
    res_shape = xy_final_shape[::-1]
    res_shapes = [tuple(res_shape)]

    while all(res_shape > tile_size):
        res_shape = res_shape // 2
        res_shapes.append(tuple(res_shape))

    return res_shapes[:-1]


def add_ome_axes_single_plane(image_np):
    """
    Reshapes np.ndarray image to match OME-zarr standard

    Parameters
    ----------
    image_np:np.ndarray
        image to which additional axes are added to meet OME-zarr standard

    Returns
    -------
    image_np:np.ndarray
        reshaped image array

    """
    return image_np.reshape((1,) * (3) + image_np.shape)


def generate_channels(channel_names, channel_colors, im_dtype):
    """
    Generate OME-zarr channels metadata

    Parameters
    ----------
    channel_names:list
    channel_colors:list
    im_dtype:np.dtype

    Returns
    -------
    channel_info:list
        list of dicts containing OME-zarr channel info
    """
    channel_info = []
    for channel_name, channel_color in zip(channel_names, channel_colors):
        channel_info.append(
            {
                "label": channel_name,
                "color": channel_color,
                "active": True,
                "window": {"start": 0, "end": int(np.iinfo(im_dtype).max)},
            }
        )
    return channel_info


def format_channel_names(channel_names, n_ch):
    """
    Format channel names and ensure number of channel names matches number of channels or default
    to C1, C2, C3, etc.

    Parameters
    ----------
    channel_names:list
        list of str that are channel names
    n_ch: int
        number of channels detected in image

    Returns
    -------
    channel_names:
        list of str that are formatted
    """
    if channel_names is None or n_ch != len(channel_names):
        channel_names = ["C{}".format(idx) for idx in range(n_ch)]
    return channel_names


def get_pyramid_info(y_size, x_size, n_ch, tile_size):
    """
    Get pyramidal info for OME-zarr output

    Parameters
    ----------
    y_size: int
        y dimension of base layer
    x_size:int
        x dimension of base layer
    n_ch:int
        number of channels in the image
    tile_size:int
        tile size of the image

    Returns
    -------
    pyr_levels
        pyramidal levels
    pyr_shapes:
        OME-zarr pyramid shapes for all levels

    """
    yx_size = np.asarray([y_size, x_size], dtype=np.int32)
    pyr_levels = calc_pyramid_levels(yx_size, tile_size)
    pyr_shapes = [(1, n_ch, 1, int(pl[0]), int(pl[1])) for pl in pyr_levels]
    return pyr_levels, pyr_shapes


def prepare_ome_zarr_group(
    zarr_store_dir,
    y_size,
    x_size,
    n_ch,
    im_dtype,
    tile_size=512,
    channel_names=None,
    channel_colors=None,
):
    """
    Prepare OME-zarr store with all meta data and channel info and initialize store

    Parameters
    ----------
    zarr_store_dir:str
        filepath to zarr store
    y_size: int
        y dimension of base layer
    x_size:int
        x dimension of base layer
    n_ch:int
        number of channels in the image
    im_dtype:np.dtype
        data type of the image
    tile_size:int
        tile size of the image
    channel_names:list
        list of str channel names
    channel_colors:
        list of hex or str channel colors

    Returns
    -------
    grp: zarr.store
        initialized store
    n_pyr_levels: int
        number of sub-resolutions
    pyr_levels: list
        shapes of sub-resolutions
    """
    store = zarr.DirectoryStore(zarr_store_dir)
    grp = zarr.group(store, overwrite=True)
    zarr_dtype = "{}{}".format(im_dtype.kind, im_dtype.itemsize)

    pyr_levels, pyr_shapes = get_pyramid_info(x_size, y_size, n_ch, tile_size)

    paths = []
    for path, pyr_shape in enumerate(pyr_levels):
        grp.create_dataset(
            str(path),
            shape=pyr_shapes[path],
            dtype=zarr_dtype,
            chunks=(1, 1, 1, tile_size, tile_size),
        )
        paths.append({"path": str(path)})

    multiscales = [
        {
            "version": "0.1",
            "datasets": paths,
        }
    ]
    grp.attrs["multiscales"] = multiscales
    n_pyr_levels = len(paths)

    channel_names = format_channel_names(channel_names, n_ch)

    n_colors = n_ch // len(COLNAME_TO_HEX) + 1
    color_palette = [*COLNAME_TO_HEX] * n_colors

    if channel_colors is None:
        channel_colors = [color_palette[idx] for idx in range(n_ch)]
    elif n_ch != len(channel_colors) and n_ch != 1:
        channel_colors = [color_palette[idx] for idx in range(n_ch)]
    elif n_ch != len(channel_colors) and n_ch == 1:
        channel_colors = ["FFFFFF"]
    else:
        channel_colors = [COLNAME_TO_HEX[ch] for ch in channel_colors]

    channel_info = generate_channels(channel_names, channel_colors, im_dtype)

    image_data = {
        'id': 1,
        'channels': channel_info,
        'rdefs': {
            'model': 'color',
        },
    }

    grp.attrs["omero"] = image_data

    return grp, n_pyr_levels, pyr_levels


def get_final_tform_info(final_transform):
    """
    Extract size and spacing information from wsireg's final transformation elastix data

    Parameters
    ----------
    final_transform:itk.Transform
        itk.Transform with added attributes containing transform data

    Returns
    -------
    y_size: int
    x_size: int
    y_spacing: float
    x_spacing: float

    """
    x_size, y_size = (
        final_transform.output_size[0],
        final_transform.output_size[1],
    )
    x_spacing, y_spacing = (
        final_transform.output_spacing[0],
        final_transform.output_spacing[1],
    )
    return int(y_size), int(x_size), float(y_spacing), float(x_spacing)


def image_to_zarr_store(zgrp, image, channel_idx, n_pyr_levels, pyr_levels):
    """
    Write image into zarr store with sub resolutions

    Parameters
    ----------
    grp: zarr.store
        initialized store
    image:sitk.Image
        image
    channel_idx:int
        which channel the image represents
    n_pyr_levels: int
    pyr_levels: list
    """
    for pyr_idx in range(n_pyr_levels):
        if pyr_idx == 0:
            image = sitk.GetArrayFromImage(image)
        else:
            resize_shape = (
                pyr_levels[pyr_idx][1],
                pyr_levels[pyr_idx][0],
            )
            image = cv2.resize(image, resize_shape, cv2.INTER_LINEAR)

        zgrp[str(pyr_idx)][
            :, channel_idx : channel_idx + 1, :, :, :
        ] = add_ome_axes_single_plane(image)


def prepare_ome_xml_str(
    y_size, x_size, n_ch, im_dtype, is_rgb, **ome_metadata
):

    omexml = OmeXml()
    if is_rgb:
        stored_shape = (1, 1, 1, y_size, x_size, n_ch)
        im_shape = (y_size, x_size, n_ch)
    else:
        stored_shape = (n_ch, 1, 1, y_size, x_size, 1)
        im_shape = (n_ch, y_size, x_size)

    omexml.addimage(
        dtype=im_dtype,
        shape=im_shape,
        # specify how the image is stored in the TIFF file
        storedshape=stored_shape,
        **ome_metadata,
    )

    return omexml.tostring().encode("utf8")


def get_final_yx_from_tform(tform_reg_im, final_transform):
    if final_transform is not None:
        y_size, x_size, y_spacing, x_spacing = get_final_tform_info(
            final_transform
        )
    else:
        y_size, x_size = (
            (tform_reg_im.im_dims[0], tform_reg_im.im_dims[1])
            if tform_reg_im.is_rgb
            else (tform_reg_im.im_dims[1], tform_reg_im.im_dims[2])
        )
        y_spacing, x_spacing = None, None
    return y_size, x_size, y_spacing, x_spacing


def transform_to_ome_zarr(tform_reg_im, output_dir, tile_size=512):

    y_size, x_size = get_final_yx_from_tform(tform_reg_im)

    n_ch = (
        tform_reg_im.im_dims[2]
        if tform_reg_im.is_rgb
        else tform_reg_im.im_dims[0]
    )
    pyr_levels, pyr_shapes = get_pyramid_info(y_size, x_size, n_ch, tile_size)
    n_pyr_levels = len(pyr_levels)
    output_file_name = str(Path(output_dir) / tform_reg_im.image_name)
    if tform_reg_im.reader in ["tifffile", "czi", "sitk"]:
        if tform_reg_im.reader == "sitk":
            full_image = sitk.ReadImage(tform_reg_im.image_filepath)
        for channel_idx in range(n_ch):
            if tform_reg_im.reader == "tifffile":
                tform_reg_im.image = tf_zarr_read_single_ch(
                    tform_reg_im.image_filepath,
                    channel_idx,
                    tform_reg_im.is_rgb,
                )
                tform_reg_im.image = np.squeeze(tform_reg_im.image)

            elif tform_reg_im.reader == "czi":
                tform_reg_im.image = czi_read_single_ch(
                    tform_reg_im.image_filepath, channel_idx
                )
                tform_reg_im.image = np.squeeze(tform_reg_im.image)
            elif tform_reg_im.reader == "sitk":
                if tform_reg_im.is_rgb:
                    tform_reg_im.image = sitk.VectorIndexSelectionCast(
                        full_image
                    )
                elif len(full_image.GetSize()) > 2:
                    tform_reg_im.image = full_image[:, :, channel_idx]
                else:
                    tform_reg_im.image = full_image

            if tform_reg_im.composite_transform is not None:
                tform_reg_im = transform_plane(tform_reg_im)
            else:
                tform_reg_im.image = sitk.GetImageFromArray(tform_reg_im.image)
            if channel_idx == 0:
                channel_names = format_channel_names(
                    tform_reg_im.channel_names, n_ch
                )
                print(f"saving to {output_file_name}.ome.zarr")

                (zgrp, n_pyr_levels, pyr_levels,) = prepare_ome_zarr_group(
                    f"{output_file_name}.ome.zarr",
                    y_size,
                    x_size,
                    n_ch,
                    tform_reg_im.im_dtype,
                    tile_size=tile_size,
                    channel_names=channel_names,
                    channel_colors=tform_reg_im.channel_colors,
                )

            image_to_zarr_store(
                zgrp, tform_reg_im.image, channel_idx, n_pyr_levels, pyr_levels
            )

    return f"{output_file_name}.ome.zarr"


def transform_plane(image, final_transform, composite_transform):

    image = sitk_transform_image(
        image,
        final_transform,
        composite_transform,
    )

    return image


def transform_to_ome_tiff(
    tform_reg_im,
    image_name,
    output_dir,
    final_transform,
    composite_transform,
    tile_size=512,
    write_pyramid=True,
):

    y_size, x_size, y_spacing, x_spacing = get_final_yx_from_tform(
        tform_reg_im, final_transform
    )

    # protect against too large tile size
    while y_size / tile_size <= 1 or x_size / tile_size <= 1:
        tile_size = tile_size // 2

    n_ch = (
        tform_reg_im.im_dims[2]
        if tform_reg_im.is_rgb
        else tform_reg_im.im_dims[0]
    )
    pyr_levels, pyr_shapes = get_pyramid_info(y_size, x_size, n_ch, tile_size)
    n_pyr_levels = len(pyr_levels)
    output_file_name = str(Path(output_dir) / image_name)
    channel_names = format_channel_names(tform_reg_im.channel_names, n_ch)

    if final_transform is not None:
        PhysicalSizeY = y_spacing
        PhysicalSizeX = x_spacing
    else:
        PhysicalSizeY = tform_reg_im.image_res
        PhysicalSizeX = tform_reg_im.image_res

    omexml = prepare_ome_xml_str(
        y_size,
        x_size,
        n_ch,
        tform_reg_im.im_dtype,
        tform_reg_im.is_rgb,
        PhysicalSizeX=PhysicalSizeX,
        PhysicalSizeY=PhysicalSizeY,
        PhysicalSizeXUnit="µm",
        PhysicalSizeYUnit="µm",
        Name=image_name,
        Channel=None if tform_reg_im.is_rgb else {"Name": channel_names},
    )
    subifds = n_pyr_levels - 1 if write_pyramid is True else None

    rgb_im_data = []

    if tform_reg_im.reader == "sitk":
        full_image = sitk.ReadImage(tform_reg_im.image_filepath)

    print(f"saving to {output_file_name}.ome.tiff")
    with TiffWriter(f"{output_file_name}.ome.tiff", bigtiff=True) as tif:
        for channel_idx in range(n_ch):
            print(f"transforming : {channel_idx}")
            if tform_reg_im.reader != "sitk":
                image = tform_reg_im.read_single_channel(channel_idx)
                image = np.squeeze(image)
                image = sitk.GetImageFromArray(image)
                image.SetSpacing(
                    (tform_reg_im.image_res, tform_reg_im.image_res)
                )
            else:
                if tform_reg_im.is_rgb:
                    image = sitk.VectorIndexSelectionCast(
                        full_image, channel_idx
                    )
                elif len(full_image.GetSize()) > 2:
                    image = full_image[:, :, channel_idx]
                else:
                    image = full_image

            if composite_transform is not None:
                image = transform_plane(
                    image, final_transform, composite_transform
                )
                print(f"transformed : {channel_idx}")

            if tform_reg_im.is_rgb:
                rgb_im_data.append(image)
            else:
                print("saving")
                if isinstance(image, sitk.Image):
                    image = sitk.GetArrayFromImage(image)

                options = dict(
                    tile=(tile_size, tile_size),
                    compression="jpeg" if tform_reg_im.is_rgb else "deflate",
                    photometric="rgb" if tform_reg_im.is_rgb else "minisblack",
                    metadata=None,
                )
                # write OME-XML to the ImageDescription tag of the first page
                description = omexml if channel_idx == 0 else None
                # write channel data
                print(f" writing channel {channel_idx} - shape: {image.shape}")
                tif.write(
                    image,
                    subifds=subifds,
                    description=description,
                    **options,
                )

                if write_pyramid:
                    for pyr_idx in range(1, n_pyr_levels):
                        resize_shape = (
                            pyr_levels[pyr_idx][0],
                            pyr_levels[pyr_idx][1],
                        )
                        image = cv2.resize(
                            image,
                            resize_shape,
                            cv2.INTER_LINEAR,
                        )
                        print(
                            f"pyr {pyr_idx} : channel {channel_idx} shape: {image.shape}"
                        )

                        tif.write(image, **options, subfiletype=1)

        if tform_reg_im.is_rgb:
            rgb_im_data = sitk.Compose(rgb_im_data)
            rgb_im_data = sitk.GetArrayFromImage(rgb_im_data)

            options = dict(
                tile=(tile_size, tile_size),
                compression="jpeg" if tform_reg_im.is_rgb else None,
                photometric="rgb" if tform_reg_im.is_rgb else "minisblack",
                metadata=None,
            )
            # write OME-XML to the ImageDescription tag of the first page
            description = omexml

            # write channel data
            tif.write(
                rgb_im_data,
                subifds=subifds,
                description=description,
                **options,
            )

            print(f"RGB shape: {rgb_im_data.shape}")
            if write_pyramid:
                for pyr_idx in range(1, n_pyr_levels):
                    resize_shape = (
                        pyr_levels[pyr_idx][0],
                        pyr_levels[pyr_idx][1],
                    )
                    rgb_im_data = cv2.resize(
                        rgb_im_data, resize_shape, cv2.INTER_LINEAR
                    )
                    print(f"pyr {pyr_idx} : RGB , shape: {rgb_im_data.shape}")

                    tif.write(rgb_im_data, **options, subfiletype=1)

    return f"{output_file_name}.ome.tiff"


def transform_to_ome_tiff_merge(
    tform_reg_im,
    image_name,
    output_dir,
    final_transform,
    composite_transform,
    tile_size=512,
    write_pyramid=True,
):

    y_size, x_size, y_spacing, x_spacing = get_final_yx_from_tform(
        tform_reg_im.images[0], final_transform[0]
    )

    # protect against too large tile size
    while y_size / tile_size <= 1 or x_size / tile_size <= 1:
        tile_size = tile_size // 2

    n_ch = tform_reg_im.n_ch
    pyr_levels, pyr_shapes = get_pyramid_info(y_size, x_size, n_ch, tile_size)
    n_pyr_levels = len(pyr_levels)
    output_file_name = str(Path(output_dir) / image_name)
    channel_names = format_channel_names(tform_reg_im.channel_names, n_ch)

    if final_transform is not None:
        PhysicalSizeY = y_spacing
        PhysicalSizeX = x_spacing
    else:
        PhysicalSizeY = tform_reg_im.image_res
        PhysicalSizeX = tform_reg_im.image_res

    omexml = prepare_ome_xml_str(
        y_size,
        x_size,
        n_ch,
        tform_reg_im.images[0].im_dtype,
        tform_reg_im.images[0].is_rgb,
        PhysicalSizeX=PhysicalSizeX,
        PhysicalSizeY=PhysicalSizeY,
        PhysicalSizeXUnit="µm",
        PhysicalSizeYUnit="µm",
        Name=image_name,
        Channel={"Name": channel_names},
    )
    subifds = n_pyr_levels - 1 if write_pyramid is True else None

    print(f"saving to {output_file_name}.ome.tiff")
    with TiffWriter(f"{output_file_name}.ome.tiff", bigtiff=True) as tif:
        for m_idx, merge_image in enumerate(tform_reg_im.images):
            merge_n_ch = merge_image.n_ch
            for channel_idx in range(merge_n_ch):
                image = merge_image.read_single_channel(channel_idx)
                image = np.squeeze(image)
                image = sitk.GetImageFromArray(image)
                image.SetSpacing(
                    (merge_image.image_res, merge_image.image_res)
                )

                if composite_transform[m_idx] is not None:
                    image = transform_plane(
                        image,
                        final_transform[m_idx],
                        composite_transform[m_idx],
                    )

                print("saving")
                if isinstance(image, sitk.Image):
                    image = sitk.GetArrayFromImage(image)

                options = dict(
                    tile=(tile_size, tile_size),
                    compression="jpeg" if merge_image.is_rgb else "deflate",
                    photometric="rgb" if merge_image.is_rgb else "minisblack",
                    metadata=None,
                )
                # write OME-XML to the ImageDescription tag of the first page
                description = omexml if channel_idx == 0 else None
                # write channel data
                print(f" writing channel {channel_idx} - shape: {image.shape}")
                tif.write(
                    image,
                    subifds=subifds,
                    description=description,
                    **options,
                )

                if write_pyramid:
                    for pyr_idx in range(1, n_pyr_levels):
                        resize_shape = (
                            pyr_levels[pyr_idx][0],
                            pyr_levels[pyr_idx][1],
                        )
                        image = cv2.resize(
                            image,
                            resize_shape,
                            cv2.INTER_LINEAR,
                        )
                        print(
                            f"pyr {pyr_idx} : channel {channel_idx} shape: {image.shape}"
                        )

                        tif.write(image, **options, subfiletype=1)

        return f"{output_file_name}.ome.tiff"


def compute_mask_to_bbox(mask, mask_padding=100):
    mask.SetSpacing((1, 1))
    mask_size = mask.GetSize()
    mask = sitk.Threshold(mask, 1, 255)
    mask = sitk.ConnectedComponent(mask)

    labstats = sitk.LabelShapeStatisticsImageFilter()
    labstats.SetBackgroundValue(0)
    labstats.ComputePerimeterOff()
    labstats.ComputeFeretDiameterOff()
    labstats.ComputeOrientedBoundingBoxOff()
    labstats.Execute(mask)

    bb_points = []
    for label in labstats.GetLabels():
        x1, y1, xw, yh = labstats.GetBoundingBox(label)
        x2, y2 = x1 + xw, y1 + yh
        lab_points = np.asarray([[x1, y1], [x2, y2]])
        bb_points.append(lab_points)

    bb_points = np.concatenate(bb_points)
    x_min = np.min(bb_points[:, 0])
    y_min = np.min(bb_points[:, 1])
    x_max = np.max(bb_points[:, 0])
    y_max = np.max(bb_points[:, 1])

    if (x_min - mask_padding) < 0:
        x_min = 0
    else:
        x_min -= mask_padding

    if (y_min - mask_padding) < 0:
        y_min = 0
    else:
        y_min -= mask_padding

    if (x_max + mask_padding) > mask_size[0]:
        x_max = mask_size[0]
    else:
        x_max += mask_padding

    if (y_max + mask_padding) > mask_size[1]:
        y_max = mask_size[1]
    else:
        y_max += mask_padding

    x_width = x_max - x_min
    y_height = y_max - y_min

    return BoundingBox(x_min, y_min, x_width, y_height)


def sitk_vect_to_gs(image):
    """
    converts simpleITK RGB image to greyscale using cv2
    Parameters
    ----------
    image
        SimpleITK image

    Returns
    -------
        Greyscale SimpleITK image
    """
    image = sitk.GetArrayFromImage(image)
    if image.shape[2] == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    elif image.shape[2] == 4:
        image = cv2.cvtColor(image, cv2.COLOR_BGRA2GRAY)

    return sitk.GetImageFromArray(image, isVector=False)


def sitk_max_int_proj(image):
    """
    Finds maximum intensity projection of multi-channel SimpleITK image

    Parameters
    ----------
    image
        multichannel impleITK image


    Returns
    -------
    SimpleITK image
    """
    # check if there are 3 dimensions (XYC)
    if len(image.GetSize()) == 3:
        return sitk.MaximumProjection(image, 2)[:, :, 0]
    else:
        print(
            'cannot perform maximum intensity project on single channel image'
        )
        return image


def sitk_inv_int(image):
    """
    inverts intensity of images for registration, useful for alignment of brightfield
    and fluorescence images
    Parameters
    ----------
    image
        SimpleITK image s

    Returns
    -------
    Intensity inverted SimpleITK image
    """
    return sitk.InvertIntensity(image)


def contrast_enhance(image):
    spacing = image.GetSpacing()
    image = sitk.GetArrayFromImage(image)
    image = cv2.convertScaleAbs(image, alpha=7, beta=1)
    image = sitk.GetImageFromArray(image)
    image.SetSpacing(spacing)
    return image


def std_prepro():
    """
    Catch all dictionary of preprocessing functions that will result in a single 2D image for
    registration
    Returns
    -------
    dictionary of processing parameters
    """
    STD_PREPRO = {
        'image_type': 'FL',
        'ch_indices': None,
        'as_uint8': True,
        'downsample': 1,
        'max_int_proj': sitk_max_int_proj,
        'inv_int': sitk_inv_int,
    }
    return STD_PREPRO


def tifffile_to_arraylike(image_filepath):
    largest_series_idx = tf_get_largest_series(image_filepath)
    image = zarr.open(
        imread(image_filepath, aszarr=True, series=largest_series_idx)
    )
    if isinstance(image, zarr.Group):
        image = image[0]

    return image, image_filepath


def ome_tifffile_to_arraylike(image_filepath):
    ome_metadata = xml2dict(TiffFile(image_filepath).ome_metadata)
    im_dims, im_dtype = get_tifffile_info(image_filepath)

    largest_series_idx = tf_get_largest_series(image_filepath)

    series_metadata = ome_metadata.get("OME").get("Image")

    if isinstance(series_metadata, list):
        series_metadata = series_metadata[largest_series_idx]

    if isinstance(series_metadata.get("Pixels").get("Channel"), list):
        samples_per_pixel = (
            series_metadata.get("Pixels")
            .get("Channel")[0]
            .get("SamplesPerPixel")
        )
    else:
        samples_per_pixel = (
            series_metadata.get("Pixels").get("Channel").get("SamplesPerPixel")
        )

    is_rgb = guess_rgb(im_dims)

    image = zarr.open(
        imread(image_filepath, aszarr=True, series=largest_series_idx)
    )

    if isinstance(image, zarr.Group):
        image = image[0]

    image = da.from_zarr(image)
    if samples_per_pixel:
        if is_rgb is False and samples_per_pixel >= 3:
            image = image.transpose(1, 2, 0)

    return image, image_filepath


#
#
# def prepare_np_image(image, preprocessing):
#     """
#     preprocess images stored as numpy arrays
#
#     Parameters
#     ----------
#     image:np.ndarray
#         image data
#     preprocessing:dict
#         whether to do some read-time pre-processing
#         - greyscale conversion (at the tile level)
#         - read individual or range of channels (at the tile level)
#
#     Returns
#     -------
#     image:sitk.Image
#         image ready for registration
#     """
#     is_rgb = guess_rgb(image.shape)
#
#     # greyscale an RGB np array
#     if preprocessing is not None:
#         if is_rgb:
#             image = grayscale(image)
#             is_rgb = False
#
#         # select channels for registration if multi-channel
#         if image.ndim > 2 and preprocessing.get("ch_indices") is not None:
#             image = np.squeeze(image[preprocessing.get("ch_indices"), :, :])
#
#     # pass nparray to sitk image
#     image = sitk.GetImageFromArray(image, isVector=is_rgb)
#
#     # sitk preprocessing
#     if (
#         preprocessing is not None
#         and preprocessing.get('as_uint8') is True
#         and image.GetPixelID() != sitk.sitkUInt8
#     ):
#         image = sitk.RescaleIntensity(image)
#         image = sitk.Cast(image, sitk.sitkUInt8)
#
#     return image
#
#
# def read_image(image_filepath, preprocessing):
#     """
#     Convenience function to read images
#
#     Parameters
#     ----------
#     image_filepath : str
#         file path to image
#     preprocessing
#         read time preprocessing dict ("as_uint8" and "ch_indices")
#     Returns
#     -------
#     image: sitk.Image
#         SimpleITK image of image data from czi, scn or other image formats read by SimpleITK
#     """
#
#     fp_ext = Path(image_filepath).suffix.lower()
#
#     if fp_ext == '.czi':
#         czi = CziRegImageReader(image_filepath)
#         scene_idx = czi.axes.index('S')
#
#         if czi.shape[scene_idx] > 1:
#             raise ValueError('multi scene czis not allowed at this time')
#
#         if preprocessing is None:
#             image = czi.asarray()
#         else:
#             image = czi.sub_asarray(
#                 channel_idx=preprocessing['ch_indices'],
#                 as_uint8=preprocessing['as_uint8'],
#             )
#         image = np.squeeze(image)
#
#         image = sitk.GetImageFromArray(image)
#
#     # find out other WSI formats read by tifffile
#     elif fp_ext in TIFFFILE_EXTS:
#         largest_series = tf_get_largest_series(image_filepath)
#
#         try:
#             image = tifffile_dask_backend(
#                 image_filepath, largest_series, preprocessing
#             )
#         except ValueError:
#             image = tifffile_zarr_backend(
#                 image_filepath, largest_series, preprocessing
#             )
#
#         if (
#             preprocessing is not None
#             and preprocessing.get('as_uint8') is True
#             and image.GetPixelID() != sitk.sitkUInt8
#         ):
#             image = sitk.RescaleIntensity(image)
#             image = sitk.Cast(image, sitk.sitkUInt8)
#     else:
#         image = sitk_backend(image_filepath, preprocessing)
#
#     return image

#
# def get_im_info(image_filepath):
#     """
#     Use CziFile and tifffile to get image dimension and other information.
#
#     Parameters
#     ----------
#     image_filepath: str
#         filepath to the image file
#
#     Returns
#     -------
#     im_dims: np.ndarray
#         image dimensions in np.ndarray
#     im_dtype: np.dtype
#         data type of the image
#     reader:str
#         whether to use "czifile" or "tifffile" to read the image
#     """
#     if Path(image_filepath).suffix.lower() == ".czi":
#         czi = CziFile(image_filepath)
#         ch_dim_idx = czi.axes.index('C')
#         y_dim_idx = czi.axes.index('Y')
#         x_dim_idx = czi.axes.index('X')
#         im_dims = np.array(czi.shape)[[ch_dim_idx, y_dim_idx, x_dim_idx]]
#         im_dtype = czi.dtype
#         reader = "czi"
#
#     elif Path(image_filepath).suffix.lower() in TIFFFILE_EXTS:
#         largest_series = tf_get_largest_series(image_filepath)
#         zarr_im = zarr.open(
#             imread(image_filepath, aszarr=True, series=largest_series)
#         )
#         zarr_im = zarr_get_base_pyr_layer(zarr_im)
#         im_dims = np.squeeze(zarr_im.shape)
#         if len(im_dims) == 2:
#             im_dims = np.concatenate([[1], im_dims])
#         im_dtype = zarr_im.dtype
#         reader = "tifffile"
#
#     else:
#         im_dims, im_dtype = get_sitk_image_info(image_filepath)
#         reader = "sitk"
#
#     return im_dims, im_dtype, reader



from enum import Enum, EnumMeta
from pathlib import Path
from typing import Dict, List, Tuple, Union

from wsireg.parameter_maps.reg_params import DEFAULT_REG_PARAM_MAPS

DEFAULT_REG_PARAM_MAPS.keys()
PATH_LIKE = Union[str, Path]


def _elx_lineparser(
    line: str,
) -> Union[Tuple[str, List[str]], Tuple[None, None]]:
    if line[0] == "(":
        params = (
            line.replace("(", "")
            .replace(")", "")
            .replace("\n", "")
            .replace('"', "")
        )
        params = params.split(" ", 1)
        k, v = params[0], params[1]
        if " " in v:
            v = v.split(" ")
            v = list(filter(lambda a: a != "", v))
        if isinstance(v, list) is False:
            v = [v]
        return k, v
    else:
        return None, None


def _read_elastix_parameter_file(
    elx_param_fp: PATH_LIKE,
) -> Dict[str, List[str]]:
    with open(
        elx_param_fp,
        "r",
    ) as f:
        lines = f.readlines()
    parameters = {}
    for line in lines:
        k, v = _elx_lineparser(line)
        if k is not None:
            parameters.update({k: v})
    return parameters


class RegModelMeta(EnumMeta):
    def __getitem__(self, name):
        try:
            return super().__getitem__(name)
        except (TypeError, KeyError):
            if isinstance(name, (str, Path)) and Path(name).exists():
                return _read_elastix_parameter_file(name)
            else:
                raise ValueError(
                    "unrecognized registration parameter, please provide"
                    "file path to elastix transform parameters or specify one of "
                    f"{[i.name for i in self]}"
                )


class RegModel(dict, Enum, metaclass=RegModelMeta):
    rigid: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["rigid"]
    affine: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["affine"]
    similarity: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["similarity"]
    nl: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["nl"]
    fi_correction: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS[
        "fi_correction"
    ]
    nl_reduced: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["nl-reduced"]
    nl_mid: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["nl-mid"]
    nl2: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["nl2"]
    rigid_expanded: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS[
        "rigid-expanded"
    ]
    rigid_test: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["rigid_test"]
    affine_test: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["affine_test"]
    similarity_test: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS[
        "similarity_test"
    ]
    nl_test: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["nl_test"]
    rigid_ams: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["rigid_ams"]
    affine_ams: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["affine_ams"]
    similarity_ams: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS[
        "similarity_ams"
    ]
    nl_ams: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["nl_ams"]
    rigid_anc: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["rigid_anc"]
    affine_anc: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["affine_anc"]
    similarity_anc: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS[
        "similarity_anc"
    ]
    nl_anc: Dict[str, List[str]] = DEFAULT_REG_PARAM_MAPS["nl_anc"]



from pathlib import Path

import numpy as np
import pytest
from tifffile import imread

from wsireg.reg_images.loader import reg_image_loader
from wsireg.reg_images.merge_reg_image import MergeRegImage
from wsireg.reg_transform_seq import RegTransformSeq
from wsireg.writers.merge_ome_tiff_writer import MergeOmeTiffWriter
from wsireg.writers.ome_tiff_writer import OmeTiffWriter


@pytest.fixture(scope="session")
def data_out_dir(tmpdir_factory):
    out_dir = tmpdir_factory.mktemp("output")
    return out_dir


@pytest.mark.usefixtures("complex_transform")
def test_OmeTiffWriter_by_plane(complex_transform, data_out_dir):
    reg_image = reg_image_loader(np.ones((1024, 1024), dtype=np.uint8), 1)
    # composite_transform, _, final_transform = prepare_wsireg_transform_data(
    #     complex_transform
    # )
    rts = RegTransformSeq(complex_transform)

    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    by_plane_fp = ometiffwriter.write_image_by_plane(
        "testimage_by_plane",
        output_dir=str(data_out_dir),
    )
    by_plane_image = reg_image_loader(by_plane_fp, 2)
    assert by_plane_image.im_dims == (1, 1024, 1024)


@pytest.mark.usefixtures("complex_transform")
def test_OmeTiffWriter_by_tile(complex_transform, data_out_dir):
    reg_image = reg_image_loader(np.ones((1024, 1024), dtype=np.uint8), 1)
    rts = RegTransformSeq(complex_transform)

    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    by_tile_fp = ometiffwriter.write_image_by_tile(
        "testimage_by_plane",
        output_dir=str(data_out_dir),
    )
    by_tile_image = reg_image_loader(by_tile_fp, 2)
    assert by_tile_image.im_dims == (1, 1024, 1024)


@pytest.mark.usefixtures("complex_transform")
def test_OmeTiffWriter_compare_tile_plane(complex_transform, data_out_dir):
    reg_image = reg_image_loader(np.ones((1024, 1024), dtype=np.uint8), 1)
    rts = RegTransformSeq(complex_transform)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    by_tile_fp = ometiffwriter.write_image_by_tile(
        "testimage_by_tile",
        output_dir=str(data_out_dir),
    )
    by_tile_image = reg_image_loader(by_tile_fp, 2)
    by_plane_fp = ometiffwriter.write_image_by_plane(
        "testimage_by_plane",
        output_dir=str(data_out_dir),
    )
    by_plane_image = reg_image_loader(by_plane_fp, 2)

    ch0_im_tile = by_tile_image.read_single_channel(0)
    ch0_im_plane = by_plane_image.read_single_channel(0)

    assert by_plane_image.im_dims == by_tile_image.im_dims
    assert np.array_equal(ch0_im_tile, ch0_im_plane)


# tests if tile padding works appropriately
@pytest.mark.usefixtures("complex_transform_larger_padded")
def test_OmeTiffWriter_by_tile_nondiv(
    complex_transform_larger_padded, data_out_dir
):
    reg_image = reg_image_loader(np.ones((1024, 1024), dtype=np.uint8), 1)

    rts = RegTransformSeq(complex_transform_larger_padded)

    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiffwriter.write_image_by_tile(
        "testimage_by_tile_nd",
        output_dir=str(data_out_dir),
    )
    by_tile_image = reg_image_loader(by_tile_fp, 2)

    assert by_tile_image.im_dims == (
        1,
        int(np.ceil(2099 / 512) * 512),
        int(np.ceil(3099 / 512) * 512),
    )


@pytest.mark.usefixtures("complex_transform_larger")
def test_OmeTiffWriter_compare_tile_plane_nondiv(
    complex_transform_larger, data_out_dir
):
    reg_image = reg_image_loader(np.ones((1024, 1024), dtype=np.uint8), 1)

    rts = RegTransformSeq(complex_transform_larger)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiffwriter.write_image_by_tile(
        "testimage_by_tile",
        output_dir=str(data_out_dir),
    )
    by_tile_image = reg_image_loader(by_tile_fp, 2)
    by_plane_fp = ometiffwriter.write_image_by_plane(
        "testimage_by_plane",
        output_dir=str(data_out_dir),
    )
    by_plane_image = reg_image_loader(by_plane_fp, 2)

    ch0_im_tile = by_tile_image.read_single_channel(0)
    ch0_im_plane = by_plane_image.read_single_channel(0)
    assert np.array_equal(
        ch0_im_tile[0:3098, 0:2098], ch0_im_plane[0:3098, 0:2098]
    )


@pytest.mark.usefixtures("simple_transform_affine")
def test_OmeTiffWriter_compare_tile_plane_rgb(
    simple_transform_affine, data_out_dir
):
    reg_image = reg_image_loader(
        np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8), 1
    )

    rts = RegTransformSeq(simple_transform_affine)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiffwriter.write_image_by_tile(
        "testimage_by_tile",
        output_dir=str(data_out_dir),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        "testimage_by_plane",
        output_dir=str(data_out_dir),
    )
    Path(by_tile_fp).as_posix()
    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile, im_plane)


@pytest.mark.usefixtures("simple_transform_affine_nl")
def test_OmeTiffWriter_compare_tile_plane_rgb_nl(
    simple_transform_affine_nl, data_out_dir
):
    reg_image = reg_image_loader(
        np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8), 1
    )

    rts = RegTransformSeq(simple_transform_affine_nl)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiffwriter.write_image_by_tile(
        "testimage_by_tile",
        output_dir=str(data_out_dir),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        "testimage_by_plane",
        output_dir=str(data_out_dir),
    )

    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile[0:2099, 0:3099, :], im_plane)


@pytest.mark.usefixtures("simple_transform_affine")
def test_OmeTiffWriter_compare_tile_plane_mc(
    simple_transform_affine, data_out_dir
):
    reg_image = reg_image_loader(
        np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8), 1
    )
    rts = RegTransformSeq(simple_transform_affine)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiffwriter.write_image_by_tile(
        "testimage_by_tile",
        output_dir=str(data_out_dir),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        "testimage_by_plane",
        output_dir=str(data_out_dir),
    )

    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile, im_plane)


@pytest.mark.usefixtures("simple_transform_affine_nl")
def test_OmeTiffWriter_compare_tile_plane_mc_nl(
    simple_transform_affine_nl, data_out_dir
):
    reg_image = reg_image_loader(
        np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8), 1
    )
    rts = RegTransformSeq(simple_transform_affine_nl)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiffwriter.write_image_by_tile(
        "testimage_by_tile",
        output_dir=str(data_out_dir),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        "testimage_by_plane",
        output_dir=str(data_out_dir),
    )

    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile[:, 0:2099, 0:3099], im_plane)


#
# @pytest.mark.usefixtures("simple_transform_affine_nl_large_output")
# def test_OmeTiffWriter_tile_large_rgb_from_disk(
#     simple_transform_affine_nl_large_output, data_out_dir
# ):
#
#     large_rgb_fp = Path(str(data_out_dir)) / "largergb.tif"
#     imwrite(
#         large_rgb_fp,
#         np.random.randint(0, 255, (16384, 16384, 3), dtype=np.uint8),
#         compression=("jpeg", 5),
#         tile=(512, 512),
#     )
#     reg_image = reg_image_loader(str(large_rgb_fp), 0.8)
#
#     (
#         composite_transform,
#         itk_transforms,
#         final_transform,
#     ) = prepare_wsireg_transform_data(simple_transform_affine_nl_large_output)
#
#     ometiffwriter = OmeTiffWriter(reg_image)
#
#     by_tile_fp = ometiffwriter.write_image_by_tile(
#         "testimage_by_tile",
#         final_transform=final_transform,
#         itk_transforms=itk_transforms,
#         composite_transform=composite_transform,
#         output_dir=str(data_out_dir),
#     )
#
#     by_plane_fp = ometiffwriter.write_image_by_plane(
#         "testimage_by_plane",
#         final_transform=final_transform,
#         composite_transform=composite_transform,
#         output_dir=str(data_out_dir),
#     )
#
#     im_tile = imread(by_tile_fp)
#     im_plane = imread(by_plane_fp)
#
#     assert np.array_equal(
#         im_tile[0:18011, 0:14020, :], im_plane[0:18011, 0:14020, :]
#     )
#
#
# @pytest.mark.usefixtures("simple_transform_affine_nl_large_output")
# def test_OmeTiffWriter_tile_large_rgb_from_disk_plane(
#     simple_transform_affine_nl_large_output, data_out_dir
# ):
#     large_rgb_fp = Path(str(data_out_dir)) / "largergb.tif"
#     imwrite(
#         large_rgb_fp,
#         np.random.randint(0, 255, (3, 16384, 16384), dtype=np.uint8),
#         compression=("jpeg", 5),
#         tile=(512, 512),
#     )
#     reg_image = reg_image_loader(str(large_rgb_fp), 0.8)
#
#     (
#         composite_transform,
#         itk_transforms,
#         final_transform,
#     ) = prepare_wsireg_transform_data(simple_transform_affine_nl_large_output)
#
#     ometiffwriter = OmeTiffWriter(reg_image)
#
#     by_plane_fp = ometiffwriter.write_image_by_plane(
#         "testimage_by_plane",
#         final_transform=final_transform,
#         composite_transform=composite_transform,
#         output_dir=str(data_out_dir),
#     )


@pytest.mark.usefixtures("simple_transform_affine_nl")
def test_MergeOmeTiffWriter_mc(simple_transform_affine_nl, data_out_dir):
    reg_image1 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8)
    reg_image2 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8)
    reg_image3 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8)

    mreg_image = MergeRegImage(
        [reg_image1, reg_image2, reg_image3],
        [1, 1, 1],
        channel_names=[["1", "2", "3"], ["1", "2", "3"], ["1", "2", "3"]],
    )
    rts = RegTransformSeq(simple_transform_affine_nl)
    merge_ometiffwriter = MergeOmeTiffWriter(
        mreg_image, reg_seq_transforms=[rts, rts, rts]
    )

    by_plane_fp = merge_ometiffwriter.merge_write_image_by_plane(
        "merge_testimage_by_plane",
        ["1", "2", "3"],
        output_dir=str(data_out_dir),
    )

    im_plane = imread(by_plane_fp)

    reg_image1_loaded = reg_image_loader(reg_image1, 1)

    ometiffwriter = OmeTiffWriter(reg_image1_loaded, reg_transform_seq=rts)

    by_plane_fp_s1 = ometiffwriter.write_image_by_plane(
        "testimage_by_plane_s1",
        output_dir=str(data_out_dir),
    )

    im_plane_s1 = imread(by_plane_fp_s1)

    reg_image2_loaded = reg_image_loader(reg_image2, 1)

    ometiffwriter = OmeTiffWriter(reg_image2_loaded, reg_transform_seq=rts)

    by_plane_fp_s2 = ometiffwriter.write_image_by_plane(
        "testimage_by_plane_s2",
        output_dir=str(data_out_dir),
    )

    im_plane_s2 = imread(by_plane_fp_s2)

    reg_image3_loaded = reg_image_loader(reg_image3, 1)

    ometiffwriter = OmeTiffWriter(reg_image3_loaded, reg_transform_seq=rts)

    by_plane_fp_s3 = ometiffwriter.write_image_by_plane(
        "testimage_by_plane_s3",
        output_dir=str(data_out_dir),
    )

    im_plane_s3 = imread(by_plane_fp_s3)

    assert im_plane.shape[0] == 9
    assert np.array_equal(im_plane[0:3, :, :], im_plane_s1)
    assert np.array_equal(im_plane[3:6, :, :], im_plane_s2)
    assert np.array_equal(im_plane[6:9, :, :], im_plane_s3)


@pytest.mark.usefixtures("simple_transform_affine_nl")
def test_MergeOmeTiffWriter_mix_merge(
    simple_transform_affine_nl, data_out_dir
):

    reg_image1 = np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint16)
    reg_image2 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint16)
    reg_image3 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8)

    mreg_image = MergeRegImage(
        [reg_image1, reg_image2, reg_image3],
        [1, 1, 1],
        channel_names=[["1", "2", "3"], ["1", "2", "3"], ["1", "2", "3"]],
    )

    rts = RegTransformSeq(simple_transform_affine_nl)
    merge_ometiffwriter = MergeOmeTiffWriter(
        mreg_image, reg_seq_transforms=[rts, rts, rts]
    )

    by_plane_fp = merge_ometiffwriter.merge_write_image_by_plane(
        "merge_testimage_by_plane_mix",
        ["1", "2", "3"],
        output_dir=str(data_out_dir),
    )

    im_plane = imread(by_plane_fp)

    assert im_plane.shape[0] == 9
